"""A module for testing many or more complicated things.

"""

import unittest
import numpy as np

from ..utility import BasicUtility
from .test_zero import random_crack_model


class Many(unittest.TestCase):
    """Many (or More complex)

    """
    def main(self):
        """Main function for module-level testing functionality.

        """
        self.test_many_inverses()
        self.test_many_thermodynamic_connection_isometric()
        self.test_many_thermodynamic_connection_isometric_0()
        self.test_many_thermodynamic_connection_isometric_b()
        self.test_many_thermodynamic_connection_isotensional()
        self.test_many_thermodynamic_connection_isotensional_0()
        self.test_many_thermodynamic_connection_isotensional_b()

    def test_many_inverses(self):
        """Function to test inverse calculation at many random points.

        """
        y = np.random.rand(88)
        x = BasicUtility().inv_fun(lambda x: x**2, y)
        self.assertTrue(np.allclose(y, x**2))

    def test_many_thermodynamic_connection_isometric_b(self):
        """Function to test the principal thermodynamic connection
        in the isometric ensemble for the isolated bending system
        at many random points.

        """
        rgn0, rgn1 = np.random.rand(2)
        model = random_crack_model()
        v = 1 + np.random.rand(88)
        self.assertTrue(
            np.allclose(
                model.Q_b_isometric(v, [rgn0, rgn1]),
                np.exp(-model.beta_A_b(
                    v, [rgn0, rgn1], ensemble='isometric', absolute=True
                ))
            )
        )
        self.assertTrue(
            np.allclose(
                np.log(1/model.Q_b_isometric(v, [rgn0, rgn1])),
                model.beta_A_b(
                    v, [rgn0, rgn1], ensemble='isometric', absolute=True
                )
            )
        )

    def test_many_thermodynamic_connection_isometric_0(self):
        """Function to test the principal thermodynamic connection
        in the isometric ensemble for the reference system
        at many random points.

        """
        rgn0, rgn1 = np.random.rand(2)
        model = random_crack_model()
        v = 1 + np.random.rand(88)
        self.assertTrue(
            np.allclose(
                model.Q_0_isometric(v, [rgn0, rgn1]),
                np.exp(-model.beta_A_0(
                    v, [rgn0, rgn1], ensemble='isometric', absolute=True
                ))
            )
        )
        self.assertTrue(
            np.allclose(
                np.log(1/model.Q_0_isometric(v, [rgn0, rgn1])),
                model.beta_A_0(
                    v, [rgn0, rgn1], ensemble='isometric', absolute=True
                )
            )
        )

    def test_many_thermodynamic_connection_isometric(self):
        """Function to test the principal thermodynamic connection
        in the isometric ensemble for the asymptotic appoximation
        of the full system at many random points.

        """
        model = random_crack_model()
        v = 1 + np.random.rand(88)
        self.assertTrue(
            np.allclose(
                model.Q_isometric(v),
                np.exp(-model.beta_A(
                    v, ensemble='isometric', absolute=True
                ))
            )
        )
        self.assertTrue(
            np.allclose(
                np.log(model.Q_isometric(v)),
                -model.beta_A(
                    v, ensemble='isometric', absolute=True
                )
            )
        )
        model = random_crack_model()
        self.assertTrue(
            np.allclose(
                model.Q_isometric(v, transition_state=True),
                np.exp(
                    -model.beta_A_abs_isometric(v, transition_state=True)
                )
            )
        )
        self.assertTrue(
            np.allclose(
                np.log(model.Q_isometric(v, transition_state=True)),
                -model.beta_A_abs_isometric(v, transition_state=True)
            )
        )

    def test_many_thermodynamic_connection_isotensional_b(self):
        """Function to test the principal thermodynamic connection
        in the isotensional ensemble for the isolated bending system
        at many random points.

        """
        rgn0, rgn1 = np.random.rand(2)
        model = random_crack_model()
        v = 1 + np.random.rand(88)
        p = model.p(v, ensemble='isotensional')
        self.assertTrue(
            np.allclose(
                model.Z_b_isotensional(p, [rgn0, rgn1]),
                np.exp(-model.beta_G_b(
                    p, [rgn0, rgn1], ensemble='isotensional', absolute=True
                ))
            )
        )
        self.assertTrue(
            np.allclose(
                np.log(1/model.Z_b_isotensional(p, [rgn0, rgn1])),
                model.beta_G_b(
                    p, [rgn0, rgn1], ensemble='isotensional', absolute=True
                )
            )
        )

    def test_many_thermodynamic_connection_isotensional_0(self):
        """Function to test the principal thermodynamic connection
        in the isotensional ensemble for the reference system
        at many random points.

        """
        rgn0, rgn1 = np.random.rand(2)
        model = random_crack_model()
        v = 1 + np.random.rand(88)
        p = model.p(v, ensemble='isotensional')
        self.assertTrue(
            np.allclose(
                model.Z_0_isotensional(p, [rgn0, rgn1]),
                np.exp(-model.beta_G_0(
                    p, [rgn0, rgn1], ensemble='isotensional', absolute=True
                ))
            )
        )
        self.assertTrue(
            np.allclose(
                np.log(1/model.Z_0_isotensional(p, [rgn0, rgn1])),
                model.beta_G_0(
                    p, [rgn0, rgn1], ensemble='isotensional', absolute=True
                )
            )
        )

    def test_many_thermodynamic_connection_isotensional(self):
        """Function to test the principal thermodynamic connection
        in the isotensional ensemble for the asymptotic appoximation
        of the full system at many random points.

        """
        model = random_crack_model()
        v = 1 + np.random.rand(88)
        p = model.p(v, ensemble='isotensional')
        self.assertTrue(
            np.allclose(
                model.Z_isotensional(p),
                np.exp(-model.beta_G(
                    p, ensemble='isotensional', absolute=True
                ))
            )
        )
        self.assertTrue(
            np.allclose(
                np.log(model.Z_isotensional(p)),
                -model.beta_G(
                    p, ensemble='isotensional', absolute=True
                )
            )
        )
        model = random_crack_model()
        self.assertTrue(
            np.allclose(
                model.Z_isotensional(p, transition_state=True),
                np.exp(
                    -model.beta_G_abs_isotensional(p, transition_state=True)
                )
            )
        )
        self.assertTrue(
            np.allclose(
                np.log(model.Z_isotensional(p, transition_state=True)),
                -model.beta_G_abs_isotensional(p, transition_state=True)
            )
        )


if __name__ == '__main__':
    unittest.main()



"""A module for testing when something is zero.

"""

import unittest
import numpy as np

from ..core import Crack
from ..utility import BasicUtility


def random_crack_model(
    N=np.random.randint(5, high=23),
    M=np.random.randint(5, high=23),
    kappa=88*(1 - np.random.rand()/2),
    alpha=2*(1 - np.random.rand()/2),
    varepsilon=88*(1 - np.random.rand()/2)
):
    """Function to produce a random crack model.

    Args:
        N (int, optional, default=random): The number of broken bonds.
        M (int, optional, default=random): The number of intact bonds.
        kappa (float, optional, default=random):
            The nondimensional bending stiffness.
        alpha (float, optional, default=random):
            The nondimensional Morse parameter.
        varepsilon (float, optional, default=random):
            The nondimensional bond energy.

    Returns:
        object: A crack model instance.

    """
    return Crack(N=N, M=M, kappa=kappa, alpha=alpha, varepsilon=varepsilon)


class Zero(unittest.TestCase):
    """Class to test for when something is zero.

    """
    def main(self):
        """Main function for module-level testing functionality.

        """
        self.test_zero_inverse()
        self.test_zero_nondimensional_energy()
        self.test_zero_nondimensional_jacobian()
        self.test_zero_nondimensional_end_force()
        self.test_zero_minimized_nondimensional_energy()
        self.test_zero_relative_nondimensional_gibbs_free_energy()
        self.test_zero_relative_nondimensional_helmholtz_free_energy()

    def test_zero_inverse(self):
        """Function to test inverse calculation at zero.

        """
        self.assertEqual(BasicUtility().inv_fun(lambda x: x, 0), 0)

    def test_zero_nondimensional_end_force(self):
        """Function to test for zero nondimensional end force.

        """
        rgn = np.random.rand()
        model = random_crack_model()
        self.assertAlmostEqual(
            model.p_b_mechanical(rgn, [rgn, rgn])[0], 0
        )
        self.assertAlmostEqual(
            model.p_0_mechanical(rgn, [rgn, rgn])[0], 0
        )
        self.assertAlmostEqual(
            model.p_mechanical(1)[0], 0
        )
        self.assertAlmostEqual(
            model.p_b_isometric(rgn, [rgn, rgn])[0], 0
        )
        self.assertAlmostEqual(
            model.p_0_isometric(rgn, [rgn, rgn])[0], 0
        )
        self.assertAlmostEqual(
            model.p_isometric(1, approach='asymptotic')[0], 0
        )
        self.assertAlmostEqual(
            model.p_b(rgn, [rgn, rgn], ensemble='isometric')[0], 0
        )
        self.assertAlmostEqual(
            model.p_b(rgn, [rgn, rgn], ensemble='isotensional')[0], 0
        )
        self.assertAlmostEqual(
            model.p_0(rgn, [rgn, rgn], ensemble='isometric')[0], 0
        )
        self.assertAlmostEqual(
            model.p_0(rgn, [rgn, rgn], ensemble='isotensional')[0], 0
        )
        self.assertAlmostEqual(
            model.p(1, ensemble='isometric', approach='asymptotic')[0], 0
        )
        self.assertAlmostEqual(
            model.p(1, ensemble='isotensional', approach='asymptotic')[0], 0
        )

    def test_zero_nondimensional_energy(self):
        """Function to test for zero nondimensional energy.

        """
        rgn = np.random.rand()
        model = random_crack_model()
        self.assertEqual(model.beta_U_00(rgn, rgn*np.ones(8)), 0)
        self.assertEqual(model.beta_U_01(rgn*np.ones(8)), 0)
        self.assertEqual(model.beta_U_0(rgn, rgn*np.ones(8)), 0)
        self.assertEqual(model.beta_U_1(np.ones(8)), 0)
        self.assertEqual(model.beta_U(1, np.ones(8)), 0)
        self.assertEqual(model.beta_Pi_00(0, rgn, rgn*np.ones(16)), 0)
        self.assertEqual(model.beta_Pi_0(0, rgn, rgn*np.ones(16)), 0)
        self.assertEqual(model.beta_Pi(0, 1, np.ones(16)), 0)

    def test_zero_nondimensional_jacobian(self):
        """Function to test for zero nondimensional Jacobian.

        """
        rgn = np.random.rand()
        model = random_crack_model()
        for j_i in model.j_U_00(rgn, rgn*np.ones(model.L)):
            self.assertAlmostEqual(j_i, 0)
        for j_i in model.j_U_0(rgn, rgn*np.ones(model.L)):
            self.assertAlmostEqual(j_i, 0)
        for j_i in model.j_U_1(np.ones(model.M)):
            self.assertAlmostEqual(j_i, 0)
        for j_i in model.j_U(1, np.ones(model.L)):
            self.assertAlmostEqual(j_i, 0)
        for j_i in model.j_Pi_00(0, rgn, rgn*np.ones(model.L)):
            self.assertAlmostEqual(j_i, 0)
        for j_i in model.j_Pi_0(0, rgn, rgn*np.ones(model.L)):
            self.assertAlmostEqual(j_i, 0)
        for j_i in model.j_Pi_1(np.ones(model.M)):
            self.assertAlmostEqual(j_i, 0)
        for j_i in model.j_Pi(0, 1, np.ones(model.L)):
            self.assertAlmostEqual(j_i, 0)

    def test_zero_minimized_nondimensional_energy(self):
        """Function to test for zero minimized nondimensional energy.

        """
        rgn0, rgn1 = np.random.rand(2)
        model = random_crack_model()
        self.assertAlmostEqual(
            model.minimize_beta_U_00(rgn0, [rgn0, rgn0])[0][0], 0
        )
        self.assertAlmostEqual(
            model.minimize_beta_U(1)[0][0], 0
        )
        self.assertAlmostEqual(
            model.minimize_beta_Pi_00(0, [rgn0, rgn1])[0][0], 0
        )
        self.assertAlmostEqual(
            model.minimize_beta_Pi(0)[0][0], 0
        )

    def test_zero_relative_nondimensional_helmholtz_free_energy(self):
        """Function to test for zero relative Helmholtz free energy.

        """
        rgn = np.random.rand()
        model = random_crack_model()
        self.assertEqual(
            model.beta_A_b_isometric(
                1, [rgn, rgn]
            ), 0
        )
        self.assertEqual(
            model.beta_A_0_isometric(
                1, [rgn, rgn]
            ), 0
        )
        self.assertEqual(
            model.beta_A_isometric(
                1, approach='asymptotic'
            ), 0
        )
        self.assertEqual(
            model.beta_A_b(
                1, [rgn, rgn], ensemble='isometric'
            )[0], 0
        )
        self.assertAlmostEqual(
            model.beta_A_b(
                rgn, [rgn, rgn], ensemble='isotensional'
            )[0], 0
        )
        self.assertEqual(
            model.beta_A_0(
                1, [rgn, rgn], ensemble='isometric'
            )[0], 0
        )
        self.assertAlmostEqual(
            model.beta_A_0(
                rgn, [rgn, rgn], ensemble='isotensional'
            )[0], 0
        )
        self.assertEqual(
            model.beta_A(
                1, ensemble='isometric', approach='asymptotic'
            )[0], 0
        )
        self.assertAlmostEqual(
            model.beta_A(
                1, ensemble='isotensional', approach='asymptotic'
            )[0], 0
        )
        self.assertEqual(
            model.beta_A(
                1, ensemble='isometric', approach='monte carlo',
                num_processes=2, num_burns=88, num_samples=88
            )[0], 0
        )
        self.assertEqual(
            model.beta_A_isometric(
                1, approach='monte carlo',
                num_processes=2, num_burns=88, num_samples=88
            )[0], 0
        )
        self.assertEqual(
            model.beta_A_isometric_monte_carlo(
                1, num_processes=2, num_burns=88, num_samples=88
            )[0], 0
        )

    def test_zero_relative_nondimensional_gibbs_free_energy(self):
        """Function to test for zero relative Gibbs free energy.

        """
        rgn0, rgn1 = np.random.rand(2)
        model = random_crack_model()
        self.assertEqual(
            model.beta_G_b_isotensional(
                0, [rgn0, rgn1]
            ), 0
        )
        self.assertEqual(
            model.beta_G_0_isotensional(
                0, [rgn0, rgn1]
            ), 0
        )
        self.assertEqual(
            model.beta_G_isotensional(
                0, approach='asymptotic'
            ), 0
        )
        self.assertAlmostEqual(
            model.beta_G_b(
                0, [1, 1], ensemble='isometric'
            )[0], 0
        )
        self.assertEqual(
            model.beta_G_b(
                0, [rgn0, rgn1], ensemble='isotensional'
            )[0], 0
        )
        self.assertAlmostEqual(
            model.beta_G_0(
                0, [1, 1], ensemble='isometric'
            )[0], 0
        )
        self.assertEqual(
            model.beta_G_0(
                0, [rgn0, rgn1], ensemble='isotensional'
            )[0], 0
        )
        self.assertAlmostEqual(
            model.beta_G(
                0, ensemble='isometric', approach='asymptotic'
            )[0], 0
        )
        self.assertEqual(
            model.beta_G(
                0, ensemble='isotensional', approach='asymptotic'
            )[0], 0
        )
        self.assertEqual(
            model.beta_G(
                0, ensemble='isotensional', approach='monte carlo',
                num_processes=2, num_burns=88, num_samples=88
            )[0], 0
        )
        self.assertEqual(
            model.beta_G_isotensional(
                0, approach='monte carlo',
                num_processes=2, num_burns=88, num_samples=88
            )[0], 0
        )
        self.assertEqual(
            model.beta_G_isotensional_monte_carlo(
                0, num_processes=2, num_burns=88, num_samples=88
            )[0], 0
        )


if __name__ == '__main__':
    unittest.main()



"""A module for testing interfaces.

"""

import unittest
import numpy as np

from ..utility import BasicUtility
from .test_zero import random_crack_model


class Interface(unittest.TestCase):
    """Class to test interfaces.

    """
    def main(self):
        """Main function for module-level testing functionality.

        """
        self.test_interface_inverse()
        self.test_interface_nondimensional_end_force()
        self.test_interface_nondimensional_end_separation()
        self.test_interface_relative_nondimensional_gibbs_free_energy()
        self.test_interface_relative_nondimensional_helmholtz_free_energy()
        self.test_interface_relative_nondimensional_reaction_rate_coefficient()

    def test_interface_inverse(self):
        """Function to test inverse interface.

        """
        self.assertEqual(
            BasicUtility().inv_fun(lambda x: x, 0).shape, (1,)
        )
        self.assertEqual(
            BasicUtility().inv_fun(lambda x: x, 0, guess=1).shape, (1,)
        )
        y = np.random.rand(8)
        self.assertEqual(
            BasicUtility().inv_fun(lambda x: x, y).shape, y.shape
        )

    def test_interface_nondimensional_end_force(self):
        """Function to test nondimensional end force interface.

        """
        p = random_crack_model().p
        p_0 = random_crack_model().p_0
        p_b = random_crack_model().p_b
        for v in 1 + 1e-1*np.random.rand(), 1 + 1e-1*np.random.rand(8):
            _ = p_b(v, np.random.rand(2))
            _ = p_b(v, np.random.rand(2), ensemble='isometric')
            _ = p_b(v, np.random.rand(2), ensemble='isotensional')
            _ = p_0(v, np.random.rand(2))
            _ = p_0(v, np.random.rand(2), ensemble='isometric')
            _ = p_0(v, np.random.rand(2), ensemble='isotensional')
            _ = p(v)
            _ = p(v, ensemble='isometric')
            _ = p(v, ensemble='isotensional')
            _ = p(v, approach='asymptotic')
            _ = p(v, ensemble='isometric', approach='asymptotic')
            _ = p(v, ensemble='isotensional', approach='asymptotic')
            _ = p(v, approach='monte carlo',
                  num_processes=1, num_burns=8, num_samples=8)
            _ = p(v, approach='monte carlo', ensemble='isometric',
                  num_processes=1, num_burns=8, num_samples=8)
            _ = p(v, approach='monte carlo', ensemble='isotensional',
                  num_processes=1, num_burns=8, num_samples=8)
        with self.assertRaises(TypeError):
            p('a')
        with self.assertRaises(TypeError):
            p_0('a', 'b')
        with self.assertRaises(TypeError):
            p_b('a', 'b')

    def test_interface_nondimensional_end_separation(self):
        """Function to test nondimensional end separation interface.

        """
        v = random_crack_model().v
        v_0 = random_crack_model().v_0
        v_b = random_crack_model().v_b
        for p in 1e-1*np.random.rand(), 1e-1*np.random.rand(8):
            _ = v_b(p, np.random.rand(2))
            _ = v_b(p, np.random.rand(2), ensemble='isometric')
            _ = v_b(p, np.random.rand(2), ensemble='isotensional')
            _ = v_0(p, np.random.rand(2))
            _ = v_0(p, np.random.rand(2), ensemble='isometric')
            _ = v_0(p, np.random.rand(2), ensemble='isotensional')
            _ = v(p)
            _ = v(p, ensemble='isometric')
            _ = v(p, ensemble='isotensional')
            _ = v(p, approach='asymptotic')
            _ = v(p, ensemble='isometric', approach='asymptotic')
            _ = v(p, ensemble='isotensional', approach='asymptotic')
            _ = v(p, approach='monte carlo',
                  num_processes=1, num_burns=8, num_samples=8)
            _ = v(p, approach='monte carlo', ensemble='isometric',
                  num_processes=1, num_burns=8, num_samples=8)
            _ = v(p, approach='monte carlo', ensemble='isotensional',
                  num_processes=1, num_burns=8, num_samples=8)
        with self.assertRaises(TypeError):
            v('a')
        with self.assertRaises(TypeError):
            v_0('a', 'b')
        with self.assertRaises(TypeError):
            v_b('a', 'b')

    def test_interface_relative_nondimensional_helmholtz_free_energy(self):
        """Function to test relative Helmholtz free energy interface.

        """
        beta_A = random_crack_model().beta_A
        beta_A_0 = random_crack_model().beta_A_0
        beta_A_b = random_crack_model().beta_A_b
        for v in 1 + 1e-1*np.random.rand(), 1 + 1e-1*np.random.rand(8):
            _ = beta_A_0(v, np.random.rand(2))
            _ = beta_A_0(v, np.random.rand(2), ensemble='isometric')
            _ = beta_A_0(v, np.random.rand(2), ensemble='isotensional')
            _ = beta_A_0(v, np.random.rand(2), absolute=False)
            _ = beta_A_0(v, np.random.rand(2), absolute=True)
            _ = beta_A_0(v, np.random.rand(2),
                         ensemble='isometric', absolute=False)
            _ = beta_A_0(v, np.random.rand(2),
                         ensemble='isotensional', absolute=False)
            _ = beta_A_0(v, np.random.rand(2),
                         ensemble='isometric', absolute=True)
            _ = beta_A_0(v, np.random.rand(2),
                         ensemble='isotensional', absolute=True)
            _ = beta_A_b(v, np.random.rand(2))
            _ = beta_A_b(v, np.random.rand(2), ensemble='isometric')
            _ = beta_A_b(v, np.random.rand(2), ensemble='isotensional')
            _ = beta_A_b(v, np.random.rand(2), absolute=False)
            _ = beta_A_b(v, np.random.rand(2), absolute=True)
            _ = beta_A_b(v, np.random.rand(2),
                         ensemble='isometric', absolute=False)
            _ = beta_A_b(v, np.random.rand(2),
                         ensemble='isotensional', absolute=False)
            _ = beta_A_b(v, np.random.rand(2),
                         ensemble='isometric', absolute=True)
            _ = beta_A_b(v, np.random.rand(2),
                         ensemble='isotensional', absolute=True)
            _ = beta_A(v)
            _ = beta_A(v, ensemble='isometric')
            _ = beta_A(v, ensemble='isotensional')
            _ = beta_A(v, approach='asymptotic')
            _ = beta_A(v, absolute=False)
            _ = beta_A(v, absolute=True)
            _ = beta_A(v, ensemble='isometric', approach='asymptotic')
            _ = beta_A(v, ensemble='isotensional', approach='asymptotic')
            _ = beta_A(v, ensemble='isometric', absolute=False)
            _ = beta_A(v, ensemble='isotensional', absolute=False)
            _ = beta_A(v, ensemble='isometric', absolute=True)
            _ = beta_A(v, ensemble='isotensional', absolute=True)
            _ = beta_A(v, absolute=False, approach='asymptotic')
            _ = beta_A(v, absolute=True, approach='asymptotic')
            _ = beta_A(v, ensemble='isometric',
                       approach='asymptotic', absolute=False)
            _ = beta_A(v, ensemble='isometric',
                       approach='asymptotic', absolute=False)
            _ = beta_A(v, ensemble='isotensional',
                       approach='asymptotic', absolute=True)
            _ = beta_A(v, ensemble='isotensional',
                       approach='asymptotic', absolute=True)
            _ = beta_A(v, approach='monte carlo',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_A(v, approach='monte carlo', ensemble='isometric',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_A(v, approach='monte carlo', ensemble='isotensional',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_A(v, approach='monte carlo',
                       absolute=False,
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_A(v, approach='monte carlo',
                       absolute=False, ensemble='isometric',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_A(v, approach='monte carlo',
                       absolute=False, ensemble='isotensional',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_A(v, approach='monte carlo',
                       absolute=True,
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_A(v, approach='monte carlo',
                       absolute=True, ensemble='isometric',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_A(v, approach='monte carlo',
                       absolute=True, ensemble='isotensional',
                       num_processes=1, num_burns=8, num_samples=8)
        with self.assertRaises(TypeError):
            beta_A('a')
        with self.assertRaises(TypeError):
            beta_A_0('a', 'b')
        with self.assertRaises(TypeError):
            beta_A_b('a', 'b')

    def test_interface_relative_nondimensional_gibbs_free_energy(self):
        """Function to test relative Gibbs free energy interface.

        """
        beta_G = random_crack_model().beta_G
        beta_G_0 = random_crack_model().beta_G_0
        beta_G_b = random_crack_model().beta_G_b
        for p in 1e-1*np.random.rand(), 1e-1*np.random.rand(8):
            _ = beta_G_0(p, np.random.rand(2))
            _ = beta_G_0(p, np.random.rand(2), ensemble='isometric')
            _ = beta_G_0(p, np.random.rand(2), ensemble='isotensional')
            _ = beta_G_0(p, np.random.rand(2), absolute=False)
            _ = beta_G_0(p, np.random.rand(2), absolute=True)
            _ = beta_G_0(p, np.random.rand(2),
                         ensemble='isometric', absolute=False)
            _ = beta_G_0(p, np.random.rand(2),
                         ensemble='isotensional', absolute=False)
            _ = beta_G_0(p, np.random.rand(2),
                         ensemble='isometric', absolute=True)
            _ = beta_G_0(p, np.random.rand(2),
                         ensemble='isotensional', absolute=True)
            _ = beta_G_b(p, np.random.rand(2))
            _ = beta_G_b(p, np.random.rand(2), ensemble='isometric')
            _ = beta_G_b(p, np.random.rand(2), ensemble='isotensional')
            _ = beta_G_b(p, np.random.rand(2), absolute=False)
            _ = beta_G_b(p, np.random.rand(2), absolute=True)
            _ = beta_G_b(p, np.random.rand(2),
                         ensemble='isometric', absolute=False)
            _ = beta_G_b(p, np.random.rand(2),
                         ensemble='isotensional', absolute=False)
            _ = beta_G_b(p, np.random.rand(2),
                         ensemble='isometric', absolute=True)
            _ = beta_G_b(p, np.random.rand(2),
                         ensemble='isotensional', absolute=True)
            _ = beta_G(p)
            _ = beta_G(p, ensemble='isometric')
            _ = beta_G(p, ensemble='isotensional')
            _ = beta_G(p, approach='asymptotic')
            _ = beta_G(p, absolute=False)
            _ = beta_G(p, absolute=True)
            _ = beta_G(p, ensemble='isometric', approach='asymptotic')
            _ = beta_G(p, ensemble='isotensional', approach='asymptotic')
            _ = beta_G(p, ensemble='isometric', absolute=False)
            _ = beta_G(p, ensemble='isotensional', absolute=False)
            _ = beta_G(p, ensemble='isometric', absolute=True)
            _ = beta_G(p, ensemble='isotensional', absolute=True)
            _ = beta_G(p, absolute=False, approach='asymptotic')
            _ = beta_G(p, absolute=True, approach='asymptotic')
            _ = beta_G(p, ensemble='isometric',
                       approach='asymptotic', absolute=False)
            _ = beta_G(p, ensemble='isometric',
                       approach='asymptotic', absolute=False)
            _ = beta_G(p, ensemble='isotensional',
                       approach='asymptotic', absolute=True)
            _ = beta_G(p, ensemble='isotensional',
                       approach='asymptotic', absolute=True)
            _ = beta_G(p, approach='monte carlo',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_G(p, approach='monte carlo', ensemble='isometric',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_G(p, approach='monte carlo', ensemble='isotensional',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_G(p, approach='monte carlo',
                       absolute=False,
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_G(p, approach='monte carlo',
                       absolute=False, ensemble='isometric',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_G(p, approach='monte carlo',
                       absolute=False, ensemble='isotensional',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_G(p, approach='monte carlo',
                       absolute=True,
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_G(p, approach='monte carlo',
                       absolute=True, ensemble='isometric',
                       num_processes=1, num_burns=8, num_samples=8)
            _ = beta_G(p, approach='monte carlo',
                       absolute=True, ensemble='isotensional',
                       num_processes=1, num_burns=8, num_samples=8)
        with self.assertRaises(TypeError):
            beta_G('a')
        with self.assertRaises(TypeError):
            beta_G_0('a', 'b')
        with self.assertRaises(TypeError):
            beta_G_b('a', 'b')

    def test_interface_relative_nondimensional_reaction_rate_coefficient(self):
        """Function to test relative nondimensional
        reaction rate coefficient interface.

        """
        k = random_crack_model().k
        k_0 = random_crack_model().k_0
        k_b = random_crack_model().k_b
        for v in 1 + 1e-1*np.random.rand(), 1 + 1e-1*np.random.rand(8):
            _ = k_0(v, np.random.rand(2))
            _ = k_0(v, np.random.rand(2), ensemble='isometric')
            _ = k_b(v, np.random.rand(2))
            _ = k_b(v, np.random.rand(2), ensemble='isometric')
            _ = k(v)
            _ = k(v, ensemble='isometric')
            _ = k(v, approach='asymptotic')
            _ = k(v, ensemble='isometric', approach='asymptotic')
            _ = k(v, approach='monte carlo',
                  num_processes=1, num_burns=8, num_samples=8)
            _ = k(v, approach='monte carlo', ensemble='isometric',
                  num_processes=1, num_burns=8, num_samples=8)
        for p in 1e-1*np.random.rand(), 1e-1*np.random.rand(8):
            _ = k_0(v, np.random.rand(2), ensemble='isotensional')
            _ = k_b(v, np.random.rand(2), ensemble='isotensional')
            _ = k(p, ensemble='isotensional')
            _ = k(p, ensemble='isotensional', approach='asymptotic')
            _ = k(p, approach='monte carlo', ensemble='isotensional',
                  num_processes=1, num_burns=8, num_samples=8)
        with self.assertRaises(TypeError):
            k('a')


if __name__ == '__main__':
    unittest.main()



"""The core module for the quasi-two-dimensional crack model.

"""

from .isotensional import CrackQ2DIsotensional


class CrackQ2D(CrackQ2DIsotensional):
    r"""The quasi-two-dimensional crack model class.

    """
    def __init__(self, **kwargs):
        """Initializes the :class:`CrackQ2D` class.

        Initialize and inherit all attributes and methods
        from a :class:`.CrackQ2DIsotensional` class instance.

        """
        CrackQ2DIsotensional.__init__(self, **kwargs)



"""A module for the crack model in the isometric ensemble.

This module consist of the class :class:`CrackIsometric` which
contains methods for computing quantities in the
isometric (constant displacement) thermodynamic ensemble.

"""

import numpy as np
import numpy.linalg as la

from .monte_carlo import CrackMonteCarlo


class CrackIsometric(CrackMonteCarlo):
    """The crack model class for the isometric ensemble.

    """
    def __init__(self, **kwargs):
        """Initializes the :class:`CrackIsometric` class.

        Initialize and inherit all attributes and methods
        from a :class:`.CrackMonteCarlo` class instance.

        """
        CrackMonteCarlo.__init__(self, **kwargs)
        rescaled_A = self.H_U_00()/self.kappa
        inv_rescaled_A = la.inv(rescaled_A)
        self.a_U_00 = np.array([
            inv_rescaled_A[0, 0], inv_rescaled_A[0, 1],
            inv_rescaled_A[1, 0], inv_rescaled_A[1, 1],
            inv_rescaled_A[0, -1], inv_rescaled_A[0, -2],
            inv_rescaled_A[1, -1], inv_rescaled_A[1, -2],
            inv_rescaled_A[-1, 0], inv_rescaled_A[-1, 1],
            inv_rescaled_A[-2, 0], inv_rescaled_A[-2, 1],
            inv_rescaled_A[-1, -1], inv_rescaled_A[-1, -2],
            inv_rescaled_A[-2, -1], inv_rescaled_A[-2, -2],
        ])
        self.det_H_U_00 = la.det(self.H_U_00())

    def Q_isometric(self, v, transition_state=False):
        r"""The nondimensional isometric partition function
        as a function of the nondimensional end separation,

        .. math::
            Q(v) = \int d\lambda \
                Q_0(v,\boldsymbol{\lambda})
                e^{-\beta U_1(\boldsymbol{\lambda})},

        approximated using the asymptotic relation

        .. math::
            Q(v) \sim
            Q_0(v, \hat{\boldsymbol{\lambda}})
            \prod_{j=1}^M \sqrt{\frac{2\pi}{\beta u''(\hat{\lambda}_j)}}
            \, e^{-\beta u(\hat{\lambda}_j)},

        which is valid for :math:`\varepsilon\gg 1`,
        where :math:`\hat{\boldsymbol{\lambda}}`
        is from minimizing :math:`\beta U`.

        Args:
            v (array_like): The nondimensional end separation.
            approach (str, optional, default='asymptotic'):
                The calculation approach.
            transition_state (bool, optional, default=False):
                Whether or not to calculate in the transition state.

        Returns:
            numpy.ndarray: The nondimensional isometric partition function.

        Example:
            Plot the rescaled nondimensional equilibrium radial distribution
            as a function of the nondimensional end separation
            in the isometric ensemble
            for an increasing nondimensional bond energy
            and compare to the limit given by the reference system:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackIsometric
                >>> v = np.linspace(1, 11, 250)
                >>> _ = plt.figure()
                >>> for varepsilon in [10, 25, 100, 1000]:
                ...     model = CrackIsometric(varepsilon=varepsilon)
                ...     r_Q = (model.Q_isometric(v)/model.Q_isometric(1)
                ...     )**(model.N**3/3/model.kappa)
                ...     _ = plt.plot(v - 1, (v - 1)**2*r_Q,
                ...                  label=r'$\varepsilon=$'+str(varepsilon))
                >>> r_Q_0 = (model.Q_0_isometric(v, [1, 1])
                ...     / model.Q_0_isometric(1, [1, 1])
                ... )**(model.N**3/3/model.kappa)
                >>> _ = plt.plot(v - 1, (v - 1)**2*r_Q_0,
                ...              'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'$(\Delta v)^2$' +
                ...     r'$\left[\frac{Q(v)}{Q(1)}\right]^{N^3/3\kappa}$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        lambda_hat = self.minimize_beta_U(
            v, transition_state=transition_state
        )[2][-self.M:]
        Q_isometric = self.Q_0_isometric(v, lambda_hat)
        Q_isometric *= np.prod(
            np.sqrt(
                2*np.pi/self.beta_u_pp(lambda_hat[transition_state:])
            ), axis=0
        )
        Q_isometric *= np.exp(-np.sum(self.beta_u(lambda_hat), axis=0))
        return Q_isometric

    def beta_A_abs_isometric(self, v, approach='asymptotic',
                             transition_state=False):
        r"""The absolute nondimensional Helmholtz free energy
        as a function of the nondimensional end separation,

        .. math::
            \beta A(v) = -\ln Q(v).

        Args:
            v (array_like): The nondimensional end separation.
            approach (str, optional, default='asymptotic'):
                The calculation approach.
            transition_state (bool, optional, default=False):
                Whether or not to calculate in the transition state.

        Returns:
            numpy.ndarray: The absolute nondimensional Helmholtz free energy.

        """
        if approach == 'asymptotic':
            lambda_hat = self.minimize_beta_U(
                v, transition_state=transition_state
            )[2][-self.M:]
            beta_A_abs_isometric = \
                self.beta_A_0_abs_isometric(v, lambda_hat)
            beta_A_abs_isometric += 0.5*np.sum(
                np.log(
                    self.beta_u_pp(lambda_hat[transition_state:])/2/np.pi
                ), axis=0
            )
            beta_A_abs_isometric += np.sum(self.beta_u(lambda_hat), axis=0)
        elif approach == 'monte carlo':
            beta_A_abs_isometric = np.nan*v
        return beta_A_abs_isometric

    def beta_A_isometric(self, v, approach='asymptotic', **kwargs):
        r"""The relative nondimensional Helmholtz free energy
        as a function of the nondimensional end separation,

        .. math::
            \beta\Delta A(v) = \beta A(v) - \beta A(1).

        Args:
            v (array_like): The nondimensional end separation.
            approach (str, optional, default='asymptotic'):
                The calculation approach.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`~.beta_A_isometric_monte_carlo`.

        Returns:
            numpy.ndarray: The relative nondimensional Helmholtz free energy.

        Example:
            Plot the rescaled nondimensional relative Helmholtz free energy
            as a function of the nondimensional end separation
            in the isometric ensemble
            for an increasing nondimensional bond energy
            and compare to the limit given by the reference system:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackIsometric
                >>> v = np.linspace(1, 11, 33)
                >>> _ = plt.figure()
                >>> for varepsilon in [10, 25, 100, 1000]:
                ...     model = CrackIsometric(varepsilon=varepsilon)
                ...     beta_A = \
                ...         model.beta_A_isometric(v, approach='asymptotic')
                ...     _ = plt.plot(v - 1, model.N**3/3/model.kappa*beta_A,
                ...                  label=r'$\varepsilon=$'+str(varepsilon))
                >>> beta_A_0 = model.beta_A_0_isometric(v, [1, 1])
                >>> _ = plt.plot(v - 1, model.N**3/3/model.kappa*beta_A_0,
                ...              'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'$(N^3/3\kappa)\Delta\beta A$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        if approach == 'asymptotic':
            beta_A_isometric = \
                self.beta_A_abs_isometric(v, approach=approach) - \
                self.beta_A_abs_isometric(1, approach=approach)
        elif approach == 'monte carlo':
            beta_A_isometric = self.beta_A_isometric_monte_carlo(v, **kwargs)
        return beta_A_isometric

    def p_isometric(self, v, approach='asymptotic', **kwargs):
        r"""The nondimensional end force
        as a function of the nondimensional end separation
        in the isometric ensemble,

        .. math::
            p(v) = \frac{\partial}{\partial v}\,\beta A(v).

        Args:
            v (array_like): The nondimensional end separation.
            approach (str, optional, default='asymptotic'):
                The calculation approach.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`~.p_isometric_monte_carlo`.

        Returns:
            numpy.ndarray: The nondimensional end force.

        Example:
            Plot the rescaled nondimensional end force
            as a function of the nondimensional end separation
            in the isometric ensemble
            for an increasing nondimensional bond energy
            and compare to the limit given by the reference system:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackIsometric
                >>> v = np.linspace(1, 11, 33)
                >>> _ = plt.figure()
                >>> for varepsilon in [10, 25, 100, 1000]:
                ...     model = CrackIsometric(varepsilon=varepsilon)
                ...     p = model.p_isometric(v, approach='asymptotic')
                ...     _ = plt.plot(v - 1, model.N**3/3/model.kappa*p,
                ...                  label=r'$\varepsilon=$'+str(varepsilon))
                >>> p_0 = model.p_0_isometric(v, [1, 1])
                >>> _ = plt.plot(v - 1, model.N**3/3/model.kappa*p_0,
                ...              'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'$(N^3/3\kappa)p$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        if approach == 'asymptotic':
            s_hat = self.minimize_beta_U(v)[2]
            lambda_hat = s_hat[-self.M:]
            p_isometric = self.p_0_isometric(v, lambda_hat)
        elif approach == 'monte carlo':
            p_isometric = self.p_isometric_monte_carlo(v, **kwargs)
        return p_isometric

    def k_isometric(self, v, approach='asymptotic', **kwargs):
        r"""The nondimensional forward reaction rate coefficient
        as a function of the nondimensional end separation
        in the isometric ensemble.

        Args:
            v (array_like): The nondimensional end separation.
            approach (str, optional, default='asymptotic'):
                The calculation approach.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`~.k_isometric_monte_carlo`.

        Returns:
            numpy.ndarray: The nondimensional forward reaction rate.

        Example:
            Plot the nondimensional forward reaction rate coefficient
            as a function of the nondimensional end separation
            in the isometric ensemble
            for an increasing nondimensional bond energy
            and compare to the limit given by the reference system:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackIsometric
                >>> v = np.linspace(1, 11, 33)
                >>> _ = plt.figure()
                >>> for varepsilon in [10, 25, 100, 1000]:
                ...     model = CrackIsometric(varepsilon=varepsilon)
                ...     _ = plt.semilogy(
                ...         v - 1, model.k_isometric(v),
                ...         label=r'$\varepsilon=$'+str(varepsilon))
                >>> _ = plt.semilogy(v - 1, model.k_0_isometric(v, [1, 1]),
                ...                  'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'$k$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        if approach == 'asymptotic':
            k_isometric = (
                self.Q_isometric(v, transition_state=True)/self.Q_isometric(v)
            ) / (
                self.Q_isometric(1, transition_state=True)/self.Q_isometric(1)
            )
        elif approach == 'monte carlo':
            k_isometric = self.k_isometric_monte_carlo(v, **kwargs)
        return k_isometric

    def k_rev_isometric(self, v):
        r"""The nondimensional reverse reaction rate coefficient
        as a function of the nondimensional end separation
        in the isometric ensemble.

        Args:
            v (array_like): The nondimensional end separation.

        Returns:
            numpy.ndarray: The nondimensional reverse reaction rate.

        """
        model_2 = CrackIsometric(
            N=self.N + 1, M=self.M - 1, kappa=self.kappa,
            alpha=self.alpha, varepsilon=self.varepsilon
        )
        return (
            self.Q_isometric(v, transition_state=True)/model_2.Q_isometric(v)
        ) / (
            self.Q_isometric(1, transition_state=True)/model_2.Q_isometric(1)
        )

    def k_net_isometric(self, v):
        r"""The nondimensional net reaction rate coefficient
        as a function of the nondimensional end separation
        in the isometric ensemble.

        Args:
            v (array_like): The nondimensional end separation.

        Returns:
            numpy.ndarray: The nondimensional net reaction rate.

        Example:
            Plot the nondimensional net reaction rate coefficient
            as a function of the nondimensional end separation
            in the isometric ensemble
            for an increasing system size:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import Crack
                >>> Np = np.logspace(-1, np.log(6)/np.log(10), 33)
                >>> _ = plt.figure()
                >>> for N in [4, 8, 16, 64]:
                ...     model = Crack(N=N)
                ...     v = model.v(Np/N, ensemble='isometric')
                ...     _ = plt.semilogy(
                ...         3*model.kappa*(v - 1)/N**2,
                ...         model.k_net(v, ensemble='isometric'),
                ...         label=r'$N=$'+str(N))
                >>> _ = plt.xlabel(r'$3\kappa\Delta v/N^2$')
                >>> _ = plt.ylabel(r'$k^\mathrm{net}/k_\mathrm{ref}$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        return self.k_isometric(v, approach='asymptotic') \
            - self.k_rev_isometric(v)

    def Q_0_isometric(self, v, lambda_):
        r"""The nondimensional isometric partition function
        as a function of the nondimensional end separation
        for the reference system,

        .. math::
            Q_0(v,\boldsymbol{\lambda}) =
            Q_b(v,\lambda_1,\lambda_2)
            e^{-\beta U_{01}(\boldsymbol{\lambda})}.

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional isometric partition function.

        Example:
            Plot the rescaled nondimensional equilibrium radial distribution
            as a function of the nondimensional end separation
            for the reference system in the isometric ensemble
            for an increasing number of broken bonds :math:`N`
            and compare to the thermodynamic limit:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackIsometric
                >>> v = np.linspace(1, 11, 250)
                >>> _ = plt.figure()
                >>> for N in [5, 10, 25, 100]:
                ...     model = CrackIsometric(N=N)
                ...     r_Q_0 = (model.Q_0_isometric(v, [1, 1])
                ...         / model.Q_0_isometric(1, [1, 1])
                ...     )**(model.N**3/3/model.kappa)
                ...     _ = plt.plot(v - 1, (v - 1)**2*r_Q_0,
                ...                  label='$N=$'+str(N))
                >>> _ = plt.plot(v - 1, (v - 1)**2*np.exp(-(v - 1)**2/2),
                ...              'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'$(\Delta v)^2$' +
                ...     r'$\left[\frac{Q_0(v)}{Q_0(1)}\right]^{N^3/3\kappa}$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        return self.Q_b_isometric(v, lambda_) * \
            np.exp(-self.beta_U_01(lambda_))

    def beta_A_0_abs_isometric(self, v, lambda_):
        r"""The absolute nondimensional Helmholtz free energy
        as a function of the nondimensional end separation
        for the reference system,

        .. math::
            \beta A_0(v,\boldsymbol{\lambda}) =
            -\ln Q_0(v,\boldsymbol{\lambda}) =
            \beta A_b(v,\lambda_1,\lambda_2)
            + \beta U_{01}(\boldsymbol{\lambda}).

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The absolute nondimensional Helmholtz free energy.

        """
        return self.beta_A_b_isometric_abs(v, lambda_) \
            + self.beta_U_01(lambda_)

    def beta_A_0_isometric(self, v, lambda_):
        r"""The relative nondimensional Helmholtz free energy
        as a function of the nondimensional end separation
        for the reference system,

        .. math::
            \beta\Delta A_0(v,\boldsymbol{\lambda}) =
            \beta A_0(v,\boldsymbol{\lambda}) -
            \beta A_0(v_0,\boldsymbol{\lambda}).

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The relative nondimensional Helmholtz free energy.

        Example:
            Plot the rescaled nondimensional relative Helmholtz free energy
            as a function of the nondimensional end separation
            for the reference system in the isometric ensemble
            for an increasing number of broken bonds :math:`N`
            and compare to the thermodynamic limit:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackIsometric
                >>> v = np.linspace(1, 11, 33)
                >>> _ = plt.figure()
                >>> for N in [5, 10, 25, 100]:
                ...     model = CrackIsometric(N=N)
                ...     beta_A_0 = model.beta_A_0_isometric(v, [1, 1])
                ...     _ = plt.plot(v - 1, N**3/3/model.kappa*beta_A_0,
                ...                  label='$N=$'+str(N))
                >>> _ = plt.plot(v - 1, (v - 1)**2/2, 'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'$(N^3/3\kappa)\beta\Delta A_0$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        return self.beta_A_0_abs_isometric(v, lambda_) \
            - self.beta_A_0_abs_isometric(1, lambda_)

    def p_0_isometric(self, v, lambda_):
        r"""The nondimensional end force
        as a function of the nondimensional end separation
        for the reference system in the isometric ensemble,

        .. math::
            p_0(v,\lambda_1,\lambda_2) =
            \frac{\partial}{\partial v}\,\beta A_0(v,\lambda_1,\lambda_2) =
            p_b(v,\boldsymbol{\lambda}).

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional end force.

        Example:
            Plot the rescaled nondimensional end force
            as a function of the nondimensional end separation
            for the reference system in the isometric ensemble
            for an increasing number of broken bonds :math:`N`
            and compare to the thermodynamic limit:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackIsometric
                >>> v = np.linspace(1, 11, 33)
                >>> _ = plt.figure()
                >>> for N in [5, 10, 25, 100]:
                ...     model = CrackIsometric(N=N)
                ...     p_0 = model.p_0_isometric(v, [1, 1])
                ...     _ = plt.plot(v - 1, N**3/3/model.kappa*p_0,
                ...                  label='$N=$'+str(N))
                >>> _ = plt.plot(v - 1, v - 1, 'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'$(N^3/3\kappa)p_0$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        return self.p_b_isometric(v, lambda_)

    def k_0_isometric(self, v, lambda_):
        """The nondimensional forward reaction rate coefficient
        as a function of the nondimensional end separation
        for the reference system in the isometric ensemble.

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional forward reaction rate.

        """
        return self.k_b_isometric(v, lambda_)

    def Q_b_isometric(self, v, lambda_):
        r"""The nondimensional isometric partition function
        as a function of the nondimensional end separation
        for the isolated bending system,

        .. math::
            Q_b(v,\lambda_1,\lambda_2) =
            \sqrt{\frac{(2\pi)^N}{\det\mathbf{H}}}
            \,e^{\frac{1}{2}\mathbf{g}^T\cdot\mathbf{H}^{-1}\cdot\mathbf{g}-f},

        where :math:`\mathbf{H}`,
        the nondimensional Hessian of the potential energy
        for the isolated bending system, and
        :math:`\mathbf{g}` and :math:`f` are

        .. math::
            \mathbf{g}(v,\lambda_1,\lambda_2) = \kappa\left(
                2v, -v, 0, \ldots, 0, -\lambda_1, 4\lambda_1 - \lambda_2
            \right)^T,

        .. math::
            f(v,\lambda_1,\lambda_2) = \frac{\kappa}{2}\left[v^2 + \lambda_1^2
                + \left(2\lambda_1 - \lambda_2\right)^2\right]
            .

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional isometric partition function.

        """
        return np.exp(-self.beta_A_b_isometric_abs(v, lambda_))

    def beta_A_b_isometric_abs(self, v, lambda_):
        r"""The absolute nondimensional Helmholtz free energy
        as a function of the nondimensional end separation
        for the isolated bending system,

        .. math::
            \beta A_b(v,\lambda_1,\lambda_2) =
            -\ln Q_b(v,\lambda_1,\lambda_2).

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The absolute nondimensional Helmholtz free energy.

        """
        rescaled_b_inv_A_b = self.np_array(v)**2*(
            4*self.a_U_00[0]
            - 2*(self.a_U_00[1] + self.a_U_00[2]) + self.a_U_00[3]
        ) + self.np_array(v)*(
            (4*lambda_[0] - lambda_[1])*(
                2*(self.a_U_00[4] + self.a_U_00[8])
                - (self.a_U_00[6] + self.a_U_00[9])
            ) + lambda_[0]*(
                self.a_U_00[7] + self.a_U_00[11]
                - 2*(self.a_U_00[5] + self.a_U_00[10])
            )
        ) + lambda_[0]**2*(
            self.a_U_00[15]
        ) + (4*lambda_[0] - lambda_[1])**2*(
            self.a_U_00[12]
        ) - lambda_[0]*(4*lambda_[0] - lambda_[1])*(
            self.a_U_00[13] + self.a_U_00[14]
        )
        return self.kappa/2*(
                self.np_array(v)**2
                + lambda_[0]**2 + (2*lambda_[0] - lambda_[1])**2
                - rescaled_b_inv_A_b
            ) + np.log(self.det_H_U_00) - self.N*np.log(2*np.pi)/2

    def beta_A_b_isometric(self, v, lambda_):
        r"""The relative nondimensional Helmholtz free energy
        as a function of the nondimensional end separation
        for the isolated bending system,

        .. math::
            \beta\Delta A_b(v,\lambda_1,\lambda_2) =
            \beta A_b(v,\lambda_1,\lambda_2) -
            \beta A_b(v_0,\lambda_1,\lambda_2).

        In the thermodynamic limit of a large number of broken bonds
        :math:`N`, this function has the asymptotic relation

        .. math::
            \beta\Delta A_b(v,\lambda_1,\lambda_2) \sim
            \frac{3\kappa}{2N^3}(v - v_0)^2 \quad\text{for }N\gg 1.

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The relative nondimensional Helmholtz free energy.

        """
        return self.beta_A_b_isometric_abs(v, lambda_) \
            - self.beta_A_b_isometric_abs(1, lambda_)

    def p_b_isometric(self, v, lambda_):
        r"""The nondimensional end force
        as a function of the nondimensional end separation
        for the isolated bending system in the isometric ensemble,

        .. math::
            p_b(v,\lambda_1,\lambda_2) =
            \frac{\partial}{\partial v}\,\beta A_b(v,\lambda_1,\lambda_2).

        In the thermodynamic limit of a large number of broken bonds
        :math:`N`, this function has the asymptotic relation

        .. math::
            p_b(v,\lambda_1,\lambda_2) \sim
            \frac{3\kappa}{N^3}(v - v_0) \quad\text{for }N\gg 1.

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional end force.

        """
        rescaled_dbdv_inv_A_b = self.np_array(v)*(
                4*self.a_U_00[0]
                - 2*(self.a_U_00[1] + self.a_U_00[2])
                + self.a_U_00[3]
            ) + lambda_[0]*(self.a_U_00[7] - 2*self.a_U_00[5]) \
            + (4*lambda_[0] - lambda_[1])*(2*self.a_U_00[4] - self.a_U_00[6])
        return self.kappa*(v - rescaled_dbdv_inv_A_b)

    def k_b_isometric(self, v, lambda_):
        """The nondimensional forward reaction rate coefficient
        as a function of the nondimensional end separation
        for the isolated bending system in the isometric ensemble.

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional forward reaction rate.

        """
        return (
            self.Q_b_isometric(v, [self.lambda_TS, lambda_[1]]) /
            self.Q_b_isometric(v, lambda_)
        ) / (
            self.Q_b_isometric(1, [self.lambda_TS, lambda_[1]]) /
            self.Q_b_isometric(1, lambda_)
        )



"""A module for the crack model treated mechanically.

This module consist of the class :class:`CrackMechanical` which
contains methods for computing quantities when treating
the crack model mechanically rather than statistical mechanically.

"""

import numpy as np
import numpy.linalg as la
from scipy.optimize import minimize

from .utility import BasicUtility


class CrackMechanical(BasicUtility):
    """The crack model class treated mechanically.

    Attributes:
        N (int): The number of broken bonds.
        M (int): The number of intact bonds.
        L (int): The total number of bonds.
        kappa (float): The nondimensional bending stiffness.
        alpha (float): The nondimensional Morse parameter.
        varepsilon (float): The nondimensional bond energy.

    """
    def __init__(self, N=8, M=8, kappa=100, alpha=1, varepsilon=100):
        """Initializes the :class:`CrackMechanical` class.

        Initialize and inherit all attributes and methods
        from a :class:`BasicUtility` class instance.

        """
        BasicUtility.__init__(self)
        self.N = N
        self.M = M
        self.L = N + M
        self.kappa = kappa
        self.alpha = alpha
        self.varepsilon = varepsilon
        self.lambda_TS = 1 + np.log(2)/alpha
        self.v_TS = minimize(
            lambda v: (
                self.beta_u(
                    self.minimize_beta_U(v)[2][-self.M]
                )
                - self.beta_u(self.lambda_TS)
            )**2,
            self.lambda_TS
        ).x[0]
        self.p_TS = self.p_mechanical(self.v_TS)[0]

    def beta_U_00(self, v, s):
        r"""The nondimensional potential energy
        of the isolated bending system,

        .. math::
            \beta U_{00}(v, \mathbf{s}) =
            \sum_{i=1}^{N+1} \frac{\kappa}{2} \left(
                s_{i-1} - 2s_i + s_{i+1}
            \right)^2,

        where :math:`s_0\equiv v`, and
        :math:`s_{N+1}\equiv\lambda_1`, and
        :math:`s_{N+2}\equiv\lambda_2`.

        Args:
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional potential energy.

        """
        vs_b = np.append([v], s[:self.N + 1])
        return self.kappa/2*np.sum(np.diff(vs_b, 2)**2)

    def beta_U_01(self, lambda_):
        r"""The nondimensional potential energy
        due to bending within the intact region,

        .. math::
            \beta U_{01}(\boldsymbol{\lambda})
            &=
            \beta U_0(v, \mathbf{s}) - \beta U_{00}(v, \mathbf{s})
            \\ &=
            \sum_{i=N+2}^{L-1} \frac{\kappa}{2} \left(
                s_{i-1} - 2s_i + s_{i+1}
            \right)^2
            \\ &=
            \sum_{j=2}^{M-1} \frac{\kappa}{2} \left(
                \lambda_{j-1} - 2\lambda_j + \lambda_{j+1}
            \right)^2,

        where :math:`s_0\equiv v`, and
        :math:`s_{N+1}\equiv\lambda_1`, and
        :math:`s_{N+2}\equiv\lambda_2`.

        Args:
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional potential energy.

        """
        return self.kappa/2*np.sum(np.diff(self.np_array(lambda_), 2)**2)

    def beta_U_0(self, v, s):
        r"""The nondimensional potential energy of the system
        due to bending, i.e. of the reference system,

        .. math::
            \beta U_0(v, \mathbf{s}) =
            \sum_{i=1}^{L-1} \frac{\kappa}{2} \left(
                s_{i-1} - 2s_i + s_{i+1}
            \right)^2,

        where :math:`s_0\equiv v`.

        Args:
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional potential energy.

        """
        vs = np.append([v], s)
        return self.kappa/2*np.sum(np.diff(vs, 2)**2)

    def beta_u(self, lambda_):
        r"""The nondimensional potential energy of a single bond
        as a function of the bond stretch, given by the
        Morse potential :footcite:`morse1929diatomic`,

        .. math::
            \beta u(\lambda) =
            \varepsilon \left[
                1 - e^{-\alpha(\lambda - 1)}
            \right]^2.

        Args:
            lambda_ (array_like): The bond stretch.

        Returns:
            numpy.ndarray: The nondimensional potential energy.

        """
        return self.varepsilon*(
            1 - np.exp(-self.alpha*(lambda_ - 1))
        )**2 + 1e88*(lambda_ > self.lambda_TS)

    def beta_u_p(self, lambda_):
        r"""The first derivative of the potential energy of a single bond
        as a function of the bond stretch,

        .. math::
            \beta u'(\lambda) =
            2\alpha\varepsilon e^{-\alpha(\lambda - 1)} \left[
                1 - e^{-\alpha(\lambda - 1)}
            \right]^2.

        Args:
            lambda_ (array_like): The bond stretch.

        Returns:
            numpy.ndarray:
                The first derivative of the nondimensional potential energy.

        """
        return 2*self.alpha*self.varepsilon * \
            np.exp(-self.alpha*(lambda_ - 1)) * \
            (1 - np.exp(-self.alpha*(lambda_ - 1)))

    def beta_u_pp(self, lambda_):
        r"""The second derivative of the potential energy of a single bond
        as a function of the bond stretch,

        .. math::
            \beta u''(\lambda) =
            2\alpha^2\varepsilon e^{-\alpha(\lambda - 1)} \left[
                2e^{-\alpha(\lambda - 1)} - 1
            \right]^2.

        Args:
            lambda_ (array_like): The bond stretch.

        Returns:
            numpy.ndarray:
                The second derivative of the nondimensional potential energy.

        """
        return 2*self.alpha**2*self.varepsilon * \
            np.exp(-self.alpha*(lambda_ - 1)) * \
            (2*np.exp(-self.alpha*(lambda_ - 1)) - 1)

    def beta_U_1(self, lambda_):
        r"""The nondimensional potential energy of the system
        due to stretching intact bonds,

        .. math::
            \beta U_1(\boldsymbol{\lambda}) =
            \sum_{j=1}^M \varepsilon\beta u(\lambda_j).

        Args:
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional potential energy.

        """
        return np.sum(self.beta_u(self.np_array(lambda_)))

    def beta_U(self, v, s):
        r"""The nondimensional potential energy of the system,

        .. math::
            \beta U(v,\mathbf{s}) =
            \beta U_0(v,\mathbf{s}) + \beta U_1(\boldsymbol{\lambda}),

        where :math:`\lambda_j\equiv s_{N+j}` and :math:`j=1,\ldots,M`.

        Args:
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional potential energy.

        """
        lambda_ = s[-self.M:]
        return self.beta_U_0(v, s) + self.beta_U_1(lambda_)

    def j_U_00(self, v, s):
        r"""The nondimensional Jacobian
        of the potential energy of the system
        for the isolated bending system,

        .. math::
            \mathbf{j}^U_00(v,\mathbf{s}) =
            \mathbf{H}^U_00 \cdot \mathbf{s} - \kappa\left(
                2v, -v, 0, \ldots, 0, -\lambda_1, 4\lambda_1-\lambda_2
            \right)^T.

        Args:
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Jacobian.

        """
        lambda_ = s[-self.M:]
        return self.H_U_00().dot(s[:self.N]) + \
            - self.kappa*np.concatenate((
                [2*v], [-v], np.zeros(self.N - 4),
                [-lambda_[0], 4*lambda_[0] - lambda_[1]]
            ))

    def j_U_0(self, v, s):
        r"""The nondimensional Jacobian
        of the potential energy of the system
        for the reference system,

        .. math::
            \mathbf{j}^U_0(v,\mathbf{s}) =
            \mathbf{H}^U_0 \cdot \mathbf{s} - \kappa\left(
                2v, -v, 0, \ldots, 0
            \right)^T

        Args:
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Jacobian.

        """
        return self.H_U_0().dot(s) + \
            - self.kappa*np.concatenate(([2*v], [-v], np.zeros(self.L - 2)))

    def j_U_1(self, lambda_):
        r"""The nondimensional Jacobian
        of the potential energy of the system
        due to stretching intact bonds,

        .. math::
            \left[\mathbf{j}^U_1(\boldsymbol{\lambda})\right]_j =
            2\alpha\varepsilon e^{-\alpha(\lambda_j-1)}\left(
                1 - e^{-\alpha(\lambda_j-1)}
            \right).

        Args:
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional Jacobian.

        """
        return np.concatenate((
            np.zeros(self.N),
            self.beta_u_p(lambda_)
        ))

    def j_U(self, v, s):
        r"""The nondimensional Jacobian
        of the potential energy of the system,

        .. math::
            \mathbf{j}^U(v,\mathbf{s}) =
            \mathbf{j}^U_0(v,\mathbf{s})
            +\mathbf{j}^U_1(\boldsymbol{\lambda}).

        where :math:`\lambda_j\equiv s_{N+j}` and :math:`j=1,\ldots,M`.

        Args:
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Jacobian.

        """
        lambda_ = s[-self.M:]
        return self.j_U_0(v, s) + self.j_U_1(lambda_)

    def j_U_TS(self, v, s):
        r"""The nondimensional Jacobian
        of the potential energy of the system
        in its transition state.

        Args:
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Jacobian.

        """
        rows_cols = np.concatenate((
            np.arange(self.N), np.arange(self.N + 1, self.L)
        ))
        return self.j_U(v, np.concatenate((
            s[:self.N], [self.lambda_TS], s[-(self.M - 1):]
        )))[rows_cols]

    def H_U_00(self):
        r"""The nondimensional Hessian
        of the potential energy
        for the isolated bending system,

        .. math::
            \left[\mathbf{H}^U_00\right]_{mn} =
            \kappa\left(6\delta^m_n - \delta^m_1\delta^n_1
            - 4\delta^{|m-n|}_1 + \delta^{|m-n|}_2\right),
            \quad m,n=1,\ldots,N.

        Returns:
            numpy.ndarray: The nondimensional Hessian.

        """
        return self.H_Pi_00()[1:, 1:]

    def H_U_0(self):
        r"""The nondimensional Hessian
        of the potential energy of the system
        due to bending,

        .. math::
            \left[\mathbf{H}^U_0\right]_{mn} =
            \left[\mathbf{H}^\pi_0\right]_{m+1,n+1},
            \quad m,n=1,\ldots,L.

        Returns:
            numpy.ndarray: The nondimensional Hessian.

        """
        return self.H_Pi_0()[1:, 1:]

    def H_U_1(self, lambda_):
        r"""The nondimensional Hessian
        of the potential energy of the system
        due to stretching intact bonds,

        .. math::
            \left[\mathbf{H}^U_1(\boldsymbol{\lambda})\right]_{mn} =
            2\alpha^2\varepsilon e^{-\alpha(\lambda_m-1)}\left(
                1 - 2e^{-\alpha(\lambda_m-1)}
            \right)\delta^m_n.

        Args:
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional Hessian.

        """
        return np.diag(
            np.concatenate((
                np.zeros(self.N),
                self.beta_u_pp(lambda_)
            ))
        )

    def H_U(self, s):
        r"""The nondimensional Hessian
        of the potential energy of the system,

        .. math::
            \mathbf{H}^U(\mathbf{s}) =
            \mathbf{H}^U_0
            +\mathbf{H}^U_1(\boldsymbol{\lambda}).

        where :math:`\lambda_j\equiv s_{N+j}` and :math:`j=1,\ldots,M`.

        Args:
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Hessian.

        """
        lambda_ = s[-self.M:]
        return self.H_U_0() + self.H_U_1(lambda_)

    def H_U_TS(self, s):
        r"""The nondimensional Hessian
        of the potential energy of the system
        in its transition state.

        Args:
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Hessian.

        """
        rows_cols = np.concatenate((
            np.arange(self.N), np.arange(self.N + 1, self.L)
        ))
        return self.H_U(np.concatenate((
                s[:self.N], [self.lambda_TS], s[-(self.M - 1):]
            )))[rows_cols, :][:, rows_cols]

    def beta_Pi_00(self, p, v, s):
        r"""The nondimensional total potential energy
        for the isolated bending system,

        .. math::
            \beta \Pi_b(p,v,\mathbf{s}) =
            \beta U_{00}(v,\mathbf{s}) - pv.

        Args:
            p (array_like): The nondimensional end force.
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        """
        return self.beta_U_00(v, s) - p*v

    def beta_Pi_0(self, p, v, s):
        r"""The nondimensional total potential energy
        of the reference system,

        .. math::
            \beta \Pi_{0}(p,v,\mathbf{s}) =
            \beta U_{0}(v,\mathbf{s}) - pv.

        Args:
            p (array_like): The nondimensional end force.
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        """
        return self.beta_U_0(v, s) - p*v

    def beta_Pi(self, p, v, s):
        r"""The nondimensional total potential energy of the system,

        .. math::
            \beta \Pi(p,v,\mathbf{s}) =
            \beta U(v,\mathbf{s}) - pv.

        Args:
            p (array_like): The nondimensional end force.
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional total potential energy.

        """
        return self.beta_U(v, s) - p*v

    def j_Pi_00(self, p, v, s):
        r"""The nondimensional Jacobian
        of the total potential energy of the system
        for the isolated bending system,

        .. math::
            \mathbf{j}^\Pi_b(p,v,\mathbf{s}) =
            \mathbf{H}^\Pi_b \cdot \left(
                v, s_1, \cdots, s_N
            \right)^T - \kappa\left(
                p/\kappa, 0, \ldots, 0, -\lambda_1, 4\lambda_1-\lambda_2
            \right)^T

        Args:
            p (array_like): The nondimensional end force.
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Jacobian.

        """
        lambda_ = s[-self.M:]
        return self.H_Pi_00().dot(np.concatenate(([v], s[:self.N]))) + \
            - self.kappa*np.concatenate((
                [p/self.kappa], np.zeros(self.N - 2),
                [-lambda_[0], 4*lambda_[0] - lambda_[1]]
            ))

    def j_Pi_0(self, p, v, s):
        r"""The nondimensional Jacobian
        of the total potential energy of the system
        due to bending,

        .. math::
            \mathbf{j}^\Pi_0(p,v,\mathbf{s}) =
            \mathbf{H}^\Pi_0 \cdot \left(
                v, s_1, \cdots, s_L
            \right)^T - \left(
                p, 0, \ldots, 0
            \right)^T

        Args:
            p (array_like): The nondimensional end force.
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Jacobian.

        """
        return self.H_Pi_0().dot(np.concatenate(([v], s))) \
            - np.concatenate(([p], np.zeros(self.L)))

    def j_Pi_1(self, lambda_):
        r"""The nondimensional Jacobian
        of the total potential energy of the system
        due to stretching intact bonds,

        .. math::
            \mathbf{j}^\Pi_1(\boldsymbol{\lambda}) =
            \mathbf{j}^U_1(\boldsymbol{\lambda}).

        Args:
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional Jacobian.

        """
        return np.concatenate((
            np.zeros(self.N + 1),
            self.beta_u_p(lambda_)
        ))

    def j_Pi(self, p, v, s):
        r"""The nondimensional Jacobian
        of the total potential energy of the system,

        .. math::
            \mathbf{j}^\Pi(p,v,\mathbf{s}) =
            \mathbf{j}^\Pi_0(p,v,\mathbf{s})
            +\mathbf{j}^\Pi_1(\boldsymbol{\lambda}).

        where :math:`\lambda_j\equiv s_{N+j}` and :math:`j=1,\ldots,M`.

        Args:
            p (array_like): The nondimensional end force.
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Jacobian.

        """
        lambda_ = s[-self.M:]
        return self.j_Pi_0(p, v, s) + self.j_Pi_1(lambda_)

    def j_Pi_TS(self, p, v, s):
        r"""The nondimensional Jacobian
        of the total potential energy of the system
        in its transition state.

        Args:
            p (array_like): The nondimensional end force.
            v (array_like): The nondimensional end separation.
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Jacobian.

        """
        rows_cols = np.concatenate((
            np.arange(self.N + 1), np.arange(self.N + 2, self.L + 1)
        ))
        return self.j_Pi(p, v, np.concatenate((
            s[:self.N], [self.lambda_TS], s[-(self.M - 1):]
        )))[rows_cols]

    def H_Pi_00(self):
        r"""The nondimensional Hessian
        of the total potential energy
        for the isolated bending system,

        .. math::
            \left[\mathbf{H}^\Pi_b\right]_{mn} =
            \kappa\left(\phantom{\delta^{|m-n|}_1}\right. &
            6\delta^m_n - 5\delta^m_1\delta^n_1
            - \delta^m_2\delta^n_2 - 4\delta^{|m-n|}_1
            \\ &
            + 2\delta^m_1\delta^n_2 + 2\delta^m_2\delta^n_1
            + \delta^{|m-n|}_2
            \left.\phantom{\delta^{|m-n|}_1}\right),
            \quad m,n=1,\ldots,N+1.

        Returns:
            numpy.ndarray: The nondimensional Hessian.

        """
        diag_0 = np.concatenate(([1, 5], 6*np.ones(self.N - 1)))
        diag_1 = np.concatenate(([-2], -4*np.ones(self.N - 1)))
        diag_2 = np.ones(self.N - 1)
        return self.kappa*(np.diag(diag_0)
                           + np.diag(diag_1, 1) + np.diag(diag_1, -1)
                           + np.diag(diag_2, 2) + np.diag(diag_2, -2))

    def H_Pi_0(self):
        r"""The nondimensional Hessian
        of the total potential energy of the system
        due to bending,

        .. math::
            \left[\mathbf{H}^\Pi_0\right]_{mn} =
            \kappa\left(\phantom{\delta^{|m-n|}_1}\right. &
            6\delta^m_n - 5\delta^m_1\delta^n_1
            - \delta^m_2\delta^n_2 - 5\delta^m_{L+1}\delta^n_{L+1}
            - \delta^m_L\delta^n_L - 4\delta^{|m-n|}_1
            \\ &
            + 2\delta^m_1\delta^n_2 + 2\delta^m_2\delta^n_1
            + 2\delta^m_{L+1}\delta^n_L + 2\delta^m_L\delta^n_{L+1}
            + \delta^{|m-n|}_2
            \left.\phantom{\delta^{|m-n|}_1}\right),
            \\ &
            \quad m,n=1,\ldots,L+1.

        Returns:
            numpy.ndarray: The nondimensional Hessian.

        """
        diag_0 = np.concatenate(([1, 5], 6*np.ones(self.L - 3), [5, 1]))
        diag_1 = np.concatenate(([-2], -4*np.ones(self.L - 2), [-2]))
        diag_2 = np.ones(self.L - 1)
        return self.kappa*(np.diag(diag_0)
                           + np.diag(diag_1, 1) + np.diag(diag_1, -1)
                           + np.diag(diag_2, 2) + np.diag(diag_2, -2))

    def H_Pi_1(self, lambda_):
        r"""The nondimensional Hessian
        of the total potential energy of the system
        due to stretching intact bonds,

        .. math::
            \left[\mathbf{H}^\Pi_1(\boldsymbol{\lambda})\right]_{mn} =
            \left[\mathbf{H}^U_1(\boldsymbol{\lambda})\right]_{mn}.

        Args:
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional Hessian.

        """
        return np.diag(
            np.concatenate((
                np.zeros(self.N + 1),
                self.beta_u_pp(lambda_)
            ))
        )

    def H_Pi(self, s):
        r"""The nondimensional Hessian
        of the total potential energy of the system,

        .. math::
            \mathbf{H}^\Pi(\mathbf{s}) =
            \mathbf{H}^\Pi_0
            +\mathbf{H}^\Pi_1(\boldsymbol{\lambda}).

        where :math:`\lambda_j\equiv s_{N+j}` and :math:`j=1,\ldots,M`.

        Args:
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Hessian.

        """
        lambda_ = s[-self.M:]
        return self.H_Pi_0() + self.H_Pi_1(lambda_)

    def H_Pi_TS(self, s):
        r"""The nondimensional Hessian
        of the total potential energy of the system
        in its transition state.

        Args:
            s (array_like): The nondimensional configuration.

        Returns:
            numpy.ndarray: The nondimensional Hessian.

        """
        rows_cols = np.concatenate((
            np.arange(self.N + 1), np.arange(self.N + 2, self.L + 1)
        ))
        return self.H_Pi(np.concatenate((
            s[:self.N], [self.lambda_TS], s[-(self.M - 1):]
        )))[rows_cols, :][:, rows_cols]

    def minimize_beta_U_00(self, v, lambda_):
        r"""Function to minimize the potential energy
        of the isolated bending system.

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            tuple:

                - (*numpy.ndarray*) -
                  The minimized nondimensional potential energy.
                - (*numpy.ndarray*) -
                  The corresponding nondimensional end separation.
                - (*numpy.ndarray*) -
                  The corresponding nondimensional positions.

        Example:
            Plot the rescaled minimized nondimensional potential energy
            as a function of the nondimensional end separation
            for the mechnically-treated isolated bending system
            for an increasing number of broken bonds :math:`N`
            and compare to the thermodynamic limit:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackMechanical
                >>> v = np.linspace(1, 11, 33)
                >>> _ = plt.figure()
                >>> for N in [5, 10, 25]:
                ...     model = CrackMechanical(N=N)
                ...     beta_U = model.minimize_beta_U_00(v, [1, 1])[0]
                ...     _ = plt.plot(v - 1, model.N**3/3/model.kappa*beta_U,
                ...                  label='$N=$'+str(N))
                ...     p_m = model.p_0_mechanical(v, [1, 1])
                ...     beta_Pi = model.minimize_beta_Pi_00(p_m, [1, 1])[0]
                ...     _ = plt.plot(v - 1,
                ...                  model.N**3/3/model.kappa*(beta_Pi+p_m*v)
                ...                  , 'k:')
                >>> _ = plt.plot(v - 1, (v - 1)**2/2, 'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'minimized $(N^3/3\kappa)\beta U$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        v = self.np_array(v)
        beta_U = np.zeros(len(v))
        p = np.zeros(len(v))
        s = np.zeros((self.N, len(v)))
        for i, v_i in enumerate(v):
            s[:, i] = la.inv(self.H_U_00()/self.kappa).dot(
                np.concatenate(([2*v_i], [-v_i], np.zeros(self.N - 4),
                                [-lambda_[0], 4*lambda_[0] - lambda_[1]])))
            p[i] = self.kappa*(v_i - (2*s[0, i] - s[1, i]))
            beta_U[i] = self.beta_U_00(v_i, s[:, i])
        return beta_U, p, s

    def minimize_beta_U(self, v, transition_state=False):
        r"""Function to minimize the potential energy of the system.

        Args:
            v (array_like): The nondimensional end separation.
            transition_state (bool, optional, default=False):
                Whether or not to fix the crack tip bond
                stretch at the transition state stretch.

        Returns:
            tuple:

                - (*numpy.ndarray*) -
                  The minimized nondimensional potential energy.
                - (*numpy.ndarray*) -
                  The corresponding nondimensional positions.

        Example:
            Plot the rescaled minimized nondimensional potential energy
            as a function of the nondimensional end separation
            for the mechnically-treated system
            for an increasing number of broken bonds :math:`N`
            and compare to the thermodynamic limit:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackMechanical
                >>> v = np.linspace(1, 11, 33)
                >>> _ = plt.figure()
                >>> for N in [5, 10, 25]:
                ...     model = CrackMechanical(N=N)
                ...     beta_U = model.minimize_beta_U(v)[0]
                ...     _ = plt.plot(v - 1, model.N**3/3/model.kappa*beta_U,
                ...                  label='$N=$'+str(N))
                ...     p_m = model.p_mechanical(v)
                ...     beta_Pi = model.minimize_beta_Pi(p_m)[0]
                ...     _ = plt.plot(v - 1,
                ...                  model.N**3/3/model.kappa*(beta_Pi+p_m*v)
                ...                  , 'k:')
                >>> _ = plt.plot(v - 1, (v - 1)**2/2, 'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'minimized $(N^3/3\kappa)\beta U$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        v = self.np_array(v)
        beta_U = np.zeros(len(v))
        p = np.zeros(len(v))
        s = np.zeros((self.L, len(v)))
        for i, v_i in enumerate(v):
            s_guess = np.concatenate((
                self.minimize_beta_U_00(
                    v_i, [1 + transition_state*(self.lambda_TS - 1), 1]
                )[2][:, 0],
                np.ones(self.M - transition_state)
            ))
            if transition_state is True:
                res = minimize(
                    lambda s: self.beta_U(v_i, np.concatenate((
                        s[:self.N], [self.lambda_TS], s[-(self.M - 1):]
                    ))),
                    s_guess,
                    method='Newton-CG',
                    jac=lambda s: self.j_U_TS(v_i, s),
                    hess=self.H_U_TS
                )
                s[:self.N, i] = res.x[:self.N]
                s[self.N, i] = self.lambda_TS
                s[-(self.M - 1):, i] = res.x[-(self.M - 1):]
            else:
                res = minimize(
                    lambda s: self.beta_U(v_i, s),
                    s_guess,
                    method='Newton-CG',
                    jac=lambda s: self.j_U(v_i, s),
                    hess=self.H_U
                )
                s[:, i] = res.x
            p[i] = self.kappa*(v_i - (2*s[0, i] - s[1, i]))
            beta_U[i] = res.fun
        return beta_U, p, s

    def minimize_beta_Pi_00(self, p, lambda_):
        r"""Function to minimize the total potential energy
        of the isolated bending system.

        Args:
            p (array_like): The nondimensional end force.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            tuple:

                - (*numpy.ndarray*) -
                  The minimized nondimensional total potential energy.
                - (*numpy.ndarray*) -
                  The corresponding nondimensional end separation.
                - (*numpy.ndarray*) -
                  The corresponding nondimensional positions.

        """
        p = self.np_array(p)
        beta_Pi = np.zeros(len(p))
        v = np.zeros(len(p))
        s = np.zeros((self.N, len(p)))
        for i, p_i in enumerate(p):
            vs = la.inv(self.H_Pi_00()/self.kappa).dot(
                np.concatenate(([p_i/self.kappa], np.zeros(self.N - 2),
                                [-lambda_[0], 4*lambda_[0] - lambda_[1]])))
            v[i] = vs[0]
            s[:, i] = vs[1:]
            beta_Pi[i] = self.beta_Pi_00(p_i, v[i], s[:, i])
        return beta_Pi, v, s

    def minimize_beta_Pi(self, p, transition_state=False):
        """Function to minimize the total potential energy of the system.

        Args:
            p (array_like): The nondimensional end force.
            transition_state (bool, optional, default=False):
                Whether or not to fix the crack tip bond
                stretch at the transition state stretch.

        Returns:
            tuple:

                - (*numpy.ndarray*) -
                  The minimized nondimensional total potential energy.
                - (*numpy.ndarray*) -
                  The corresponding nondimensional end separation.
                - (*numpy.ndarray*) -
                  The corresponding nondimensional positions.

        """
        p = self.np_array(p)
        beta_Pi = np.zeros(len(p))
        v = np.zeros(len(p))
        s = np.zeros((self.L, len(p)))
        for i, p_i in enumerate(p):
            v_guess, s_guess_0 = self.minimize_beta_Pi_00(
                p_i, [1 + transition_state*(self.lambda_TS - 1), 1]
            )[1:]
            s_guess = np.concatenate((
                v_guess,
                s_guess_0[:, 0],
                np.ones(self.M - transition_state)
            ))
            if transition_state is True:
                res = minimize(
                    lambda vs: self.beta_Pi(p_i, vs[0], np.concatenate((
                        vs[1:self.N + 1], [self.lambda_TS], vs[-(self.M - 1):]
                    ))),
                    s_guess,
                    method='Newton-CG',
                    jac=lambda vs: self.j_Pi_TS(p_i, vs[0], vs[1:]),
                    hess=lambda vs: self.H_Pi_TS(vs[1:])
                )
                s[:self.N, i] = res.x[1:self.N + 1]
                s[self.N, i] = self.lambda_TS
                s[-(self.M - 1):, i] = res.x[-(self.M - 1):]
            else:
                res = minimize(
                    lambda vs: self.beta_Pi(p_i, vs[0], vs[1:]),
                    s_guess,
                    method='Newton-CG',
                    jac=lambda vs: self.j_Pi(p_i, vs[0], vs[1:]),
                    hess=lambda vs: self.H_Pi(vs[1:])
                )
                s[:, i] = res.x[1:]
            v[i] = res.x[0]
            beta_Pi[i] = res.fun
        return beta_Pi, v, s

    def v_mechanical(self, p):
        r"""The nondimensional end separation
        as a function of the nondimensional end force
        for the mechnically-treated system,
        calculated by minimizing the total potential energy.

        Args:
            p (array_like): The nondimensional end force.

        Returns:
            numpy.ndarray: The nondimensional end separation.

        Example:
            Plot the nondimensional end separation as a function of the
            rescaled nondimensional end force
            for the mechnically-treated system
            for an increasing nondimensional bond energy :math:`\varepsilon`
            and compare to the limit given by
            the mechanically-treated reference system:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackMechanical
                >>> v = np.linspace(1, 11, 33)
                >>> _ = plt.figure()
                >>> for varepsilon in [50, 100, 500]:
                ...     model = CrackMechanical(varepsilon=varepsilon)
                ...     r_p = np.linspace(0, 10, 33)
                ...     p = 3*model.kappa/model.N**3*r_p
                ...     v_m = model.v_mechanical(p)
                ...     _ = plt.plot(v_m - 1, r_p,
                ...                  label=r'$\varepsilon=$'+str(varepsilon))
                ...     p_m = model.p_mechanical(v_m)
                ...     r_p_m = model.N**3/3/model.kappa*p_m
                ...     _ = plt.plot(v_m - 1, r_p_m, 'k:')
                >>> _ = plt.plot(model.v_0_mechanical(p, [1, 1]) - 1,
                ...              r_p, 'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v_m$')
                >>> _ = plt.ylabel(r'$(N^3/3\kappa)p$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        return self.minimize_beta_Pi(p)[1]

    def v_0_mechanical(self, p, lambda_):
        r"""The nondimensional end separation
        as a function of the nondimensional end force
        for the mechnically-treated reference system,
        calculated by minimizing the total potential energy.

        Args:
            p (array_like): The nondimensional end force.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional end separation.

        """
        return self.v_b_mechanical(p, lambda_)

    def v_b_mechanical(self, p, lambda_):
        r"""The nondimensional end separation
        as a function of the nondimensional end force
        for the mechnically-treated isolated bending system,
        calculated by minimizing the total potential energy.

        Args:
            p (array_like): The nondimensional end force.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional end separation.

        """
        return self.minimize_beta_Pi_00(p, lambda_)[1]

    def p_mechanical(self, v):
        r"""The nondimensional end force
        as a function of the nondimensional end separation
        for the mechnically-treated system,
        calculated by minimizing the potential energy.

        Args:
            v (array_like): The nondimensional end separation.

        Returns:
            numpy.ndarray: The nondimensional end force.

        Example:
            Plot the rescaled nondimensional end force
            as a function of the nondimensional end separation
            for the mechnically-treated system
            for an increasing number of broken bonds :math:`N`
            and compare to the thermodynamic limit:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackMechanical
                >>> v = np.linspace(1, 11, 33)
                >>> _ = plt.figure()
                >>> for N in [5, 10, 25]:
                ...     model = CrackMechanical(N=N)
                ...     p_m = model.p_mechanical(v)
                ...     r_p_m = model.N**3/3/model.kappa*p_m
                ...     _ = plt.plot(v - 1, r_p_m, label='$N=$'+str(N))
                ...     v_m = model.v_mechanical(p_m)
                ...     _ = plt.plot(v_m - 1, r_p_m, 'k:')
                >>> _ = plt.plot(v - 1, v - 1, 'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'$(N^3/3\kappa)p_m$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        return self.minimize_beta_U(v)[1]

    def p_0_mechanical(self, v, lambda_):
        r"""The nondimensional end force
        as a function of the nondimensional end separation
        for the mechnically-treated reference system,
        calculated by minimizing the potential energy.

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional end force.

        Example:
            Plot the rescaled nondimensional end force
            as a function of the nondimensional end separation
            for the mechnically-treated reference system
            for an increasing number of broken bonds :math:`N`
            and compare to the thermodynamic limit:

            .. plot::

                >>> import numpy as np
                >>> import matplotlib.pyplot as plt
                >>> from statmechcrack import CrackMechanical
                >>> v = np.linspace(1, 11, 33)
                >>> _ = plt.figure()
                >>> for N in [5, 10, 25]:
                ...     model = CrackMechanical(N=N)
                ...     p_m = model.p_0_mechanical(v, [1, 1])
                ...     r_p_m = model.N**3/3/model.kappa*p_m
                ...     _ = plt.plot(v - 1, r_p_m, label='$N=$'+str(N))
                ...     v_m = model.v_0_mechanical(p_m, [1, 1])
                ...     _ = plt.plot(v_m - 1, r_p_m, 'k:')
                >>> _ = plt.plot(v - 1, v - 1, 'k--', label='limit')
                >>> _ = plt.xlabel(r'$\Delta v$')
                >>> _ = plt.ylabel(r'$(N^3/3\kappa)p_m$')
                >>> _ = plt.legend()
                >>> plt.show()

        """
        return self.p_b_mechanical(v, lambda_)

    def p_b_mechanical(self, v, lambda_):
        r"""The nondimensional end force
        as a function of the nondimensional end separation
        for the mechnically-treated reference system,
        calculated by minimizing the potential energy.

        Args:
            v (array_like): The nondimensional end separation.
            lambda_ (array_like): The intact bond stretches.

        Returns:
            numpy.ndarray: The nondimensional end force.

        """
        return self.minimize_beta_U_00(v, lambda_)[1]



"""A module for the Metropolis-Hastings Markov chain Monte Carlo method.

This module consist of the class :class:`CrackMonteCarlo` which contains
methods to compute crack model quantities using the
Metropolis-Hastings Markov chain Monte Carlo
method :footcite:`haile1992molecular`.

"""

from sys import platform
import numpy as np
import numpy.linalg as la

from .mechanical import CrackMechanical

if platform in ('linux', 'linux2'):
    import multiprocessing as mp


class CrackMonteCarlo(CrackMechanical):
    """The Metropolis-Hastings Markov chain Monte Carlo class.

    """
    def __init__(self, **kwargs):
        """Initializes the :class:`CrackMonteCarlo` class.

        Initialize and inherit all attributes and methods
        from a :class:`.CrackMechanical` class instance.

        """
        CrackMechanical.__init__(self, **kwargs)
        self.beta_E = None

    def trial_config(self, prev_config, cov_config=1e-2):
        """Generates trial configurations given a previous configuration.

        Args:
            prev_config (numpy.ndarray): The previous configuration.
            cov_config (float, optional, default=1e-2): The covariance.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            numpy.ndarray: The generated trial configuration.

        """
        return prev_config + np.random.normal(0, cov_config, len(prev_config))

    def mh_next_config(self, prev_config, urng=np.random.uniform, **kwargs):
        """Generates accepted configurations given a previous configuration.

        This function returns an accepted configuration given a
        previous configuration. Trial configurations are generated and
        repeatedly rejected until accepted as the next configuration
        based on the Metropolis-Hastings algorithm.

        Args:
            prev_config (numpy.ndarray): The previous configuration.
            urng (object, optional, default=np.random.uniform):
                The uniform random number generator.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`trial_config`.

        Returns:
            numpy.ndarray: The accepted (next) configuration.

        """
        next_config = None
        while next_config is None:
            trial_config = self.trial_config(prev_config)
            delta_beta_E = self.beta_E(trial_config) - self.beta_E(prev_config)
            if (delta_beta_E < 0) or (np.exp(-delta_beta_E) > urng()):
                next_config = trial_config
        return next_config

    def mh_next_config_append(self, prev_configs, **kwargs):
        """Append next configuration onto a history of previous configurations.

        This method calls :meth:`mh_next_config` and appends the resulting next
        configuration onto a given history of previous configurations.

        Args:
            prev_configs (numpy.ndarray): The previous
                history of configurations.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`mh_next_config`.

        Returns:
            numpy.ndarray: The updated history of configurations.

        """
        return np.append(prev_configs, np.array(
            [self.mh_next_config(prev_configs[-1], **kwargs)]), axis=0)

    def burn_in(self, init_config, num_burns=10000, tol=np.inf, **kwargs):
        """Burn-in (throw away samples) until in the effective sampling region.

        This method runs a Monte Carlo calculation without utilizing samples,
        instead throwing them away until a desired number have been thrown away
        or a desired tolerance has been reached.
        This is typically done to obtain a burned-in configuration to use as
        the initial configuration for an actual Monte Carlo calculation.

        Args:
            init_config (numpy.ndarray):
                The initial configuration.
            num_burns (int, optional, default=10000):
                The number of samples to burn.
            tol (float, optional, default=np.inf):
                The burn-in tolerance.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`mh_next_config_append`.

        Returns:
            numpy.ndarray: The final, burned-in configuration.

        """
        config_history = np.array([init_config])
        prev_rmc = init_config
        n_res = tol + 1
        while n_res > tol or len(config_history) < num_burns:
            config_history = self.mh_next_config_append(
                config_history, **kwargs
            )
            next_rmc = np.mean(config_history, axis=0)
            n_res = la.norm(next_rmc - prev_rmc)/la.norm(prev_rmc)
            prev_rmc = next_rmc
        return config_history[-1]

    def parallel_calculation(self, serial_fun, init_config, **kwargs):
        """Monte Carlo calculation averaged over several parallel processes.
        This method performs several Monte Carlo calculations in parallel
        and returns the average result from all processes.
        Each Monte Carlo calculation is performed using ``serial_fun()``.
        The default is to utilize all available processors.

        Args:
            serial_fun (function):
                The function for a single Monte Carlo calculation.
            init_config (np.ndarray):
                The initial configuration for the burn-in process.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`burn_in` and ``serial_fun()``.

        Returns:
            float: The process-averaged result.

        Note:
            To avoid issues with ``multiprocessing`` and MacOS/Windows,
            ``num_processes`` is set to 1 for anything but Linux.

        """
        burned_in_config = self.burn_in(init_config, **kwargs)
        num_processes = 1
        if platform in ('linux', 'linux2'):
            num_processes = kwargs.get('num_processes', mp.cpu_count())
        if num_processes > 1:
            output = mp.Queue()

            def fun(seed, output):
                output.put(
                    serial_fun(
                        burned_in_config,
                        urng=np.random.RandomState(seed).random, **kwargs
                    )
                )

            processes = [
                mp.Process(target=fun, args=(seed, output))
                for seed in np.random.randint(88, size=num_processes)
            ]
            for p in processes:
                p.start()
            for p in processes:
                p.join()
            process_results = [output.get() for p in processes]
            return np.mean(process_results)
        return serial_fun(burned_in_config, **kwargs)

    def p_isometric_monte_carlo_serial(self, v, init_config,
                                       num_samples=1000000, **kwargs):
        """Serial calculation for :meth:`p_isometric_monte_carlo`.

        Args:
            v (array_like):
                The nondimensional end separation.
            init_config (np.ndarray):
                The initial configuration, typically already burned-in.
            num_samples (int, optional, default=1000000):
                The number of samples to use.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`mh_next_config`.

        Returns:
            float: The nondimensional end force.

        """
        fun = 0
        config = init_config
        for counter in range(1, 1 + num_samples):
            fun_next = self.p_0_isometric(v, config)*np.exp(
                -self.beta_A_b_isometric(v, config)
            )
            fun += (fun_next - fun)/counter
            config = self.mh_next_config(config, **kwargs)
        return fun

    def p_isometric_monte_carlo(self, v, **kwargs):
        r"""The nondimensional end force
        as a function of the nondimensional end separation
        in the isometric ensemble, using a
        Metropolis-Hastings Markov chain Monte Carlo calculation
        of the ensemble average

        .. math::
            p(v) =
            e^{\beta\Delta A} \left\langle
                p_0 \, e^{-\beta\Delta A_0}
            \right\rangle_\star.

        Args:
            v (array_like): The nondimensional end separation.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`parallel_calculation`.

        Returns:
            float: The nondimensional end force.

        """
        v = self.np_array(v)
        ensemble_average_fun = np.zeros(v.shape)
        for i, v_i in enumerate(v):
            self.beta_E = lambda lambda_: self.beta_U_1(lambda_) + \
                self.beta_A_0_abs_isometric(1, lambda_)

            def serial_fun(init_config, **kwargs):
                return self.p_isometric_monte_carlo_serial(
                        v_i, init_config, **kwargs
                    )

            ensemble_average_fun[i] = self.parallel_calculation(
                serial_fun,
                self.minimize_beta_U(v_i)[2][-self.M:, 0],
                **kwargs
            )
        beta_Delta_A = self.beta_A_isometric_monte_carlo(v, **kwargs)
        return np.exp(beta_Delta_A)*ensemble_average_fun

    def v_isotensional_monte_carlo_serial(self, p, init_config,
                                          num_samples=1000000, **kwargs):
        """Serial calculation for :meth:`v_isotensional_monte_carlo`.

        Args:
            init_config (np.ndarray):
                The initial configuration, typically already burned-in.
            num_samples (int, optional, default=1000000):
                The number of samples to use.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`mh_next_config`.

        Returns:
            float: The nondimensional end separation.

        """
        fun = 0
        config = init_config
        for counter in range(1, 1 + num_samples):
            fun_next = self.v_0_isotensional(p, config)*np.exp(
                -self.beta_G_b_isotensional(p, config)
            )
            fun += (fun_next - fun)/counter
            config = self.mh_next_config(config, **kwargs)
        return fun

    def v_isotensional_monte_carlo(self, p, **kwargs):
        r"""The nondimensional end separation
        as a function of the nondimensional end force
        in the isotensional ensemble, using a
        Metropolis-Hastings Markov chain Monte Carlo calculation
        of the ensemble average

        .. math::
            v(p) =
            e^{\beta\Delta G} \left\langle
                v_0 \, e^{-\beta\Delta G_0}
            \right\rangle_\star.

        Args:
            p (array_like): The nondimensional end force.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`parallel_calculation`.

        Returns:
            float: The nondimensional end separation.

        """
        p = self.np_array(p)
        ensemble_average_fun = np.zeros(p.shape)
        for i, p_i in enumerate(p):
            self.beta_E = lambda lambda_: self.beta_U_1(lambda_) + \
                self.beta_G_0_abs_isotensional(0, lambda_)

            def serial_fun(init_config, **kwargs):
                return self.v_isotensional_monte_carlo_serial(
                        p_i, init_config, **kwargs
                    )

            ensemble_average_fun[i] = self.parallel_calculation(
                serial_fun,
                self.minimize_beta_Pi(p_i)[2][-self.M:, 0],
                **kwargs
            )
        beta_Delta_G = self.beta_G_isotensional_monte_carlo(p, **kwargs)
        return np.exp(beta_Delta_G)*ensemble_average_fun

    def beta_A_isometric_monte_carlo_serial(self, v, init_config,
                                            num_samples=1000000, **kwargs):
        """Serial calculation for :meth:`beta_A_isometric_monte_carlo`.

        Args:
            v (array_like):
                The nondimensional end separation.
            init_config (np.ndarray):
                The initial configuration, typically already burned-in.
            num_samples (int, optional, default=1000000):
                The number of samples to use.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`mh_next_config`.

        Returns:
            numpy.ndarray: The relative nondimensional Helmholtz free energy.

        """
        exp_neg_beta_Delta_A = 0
        config = init_config
        for counter in range(1, 1 + num_samples):
            exp_neg_beta_Delta_A_next = np.exp(
                -self.beta_A_b_isometric(v, config)
            )
            exp_neg_beta_Delta_A += (
                exp_neg_beta_Delta_A_next - exp_neg_beta_Delta_A
            )/counter
            config = self.mh_next_config(config, **kwargs)
        return np.log(1/exp_neg_beta_Delta_A)

    def beta_A_isometric_monte_carlo(self, v, **kwargs):
        r"""The relative nondimensional Helmholtz free energy
        as a function of the nondimensional end separation
        in the isometric ensemble, using a
        Metropolis-Hastings Markov chain Monte Carlo calculation
        of the ensemble average

        .. math::
            \beta\Delta A(v) =
            -\ln\left\langle e^{-\beta\Delta A_0} \right\rangle_\star,

        where the relative free energy of the reference system is

        .. math::
            \Delta A_0\equiv
            A_0(v,\boldsymbol{\lambda})-A_0(1,\boldsymbol{\lambda}),

        and where the isometric ensemble average here is

        .. math::
            \left\langle \phi \right\rangle_\star \equiv
            \frac{
                \int d\lambda \ e^{-\beta A_\star(\boldsymbol{\lambda})}
                    \, \phi(\boldsymbol{\lambda})
            }{
                \int d\lambda \ e^{-\beta A_\star(\boldsymbol{\lambda})}
            }
            ,

        where the free energy measure here is

        .. math::
            A_\star(\boldsymbol{\lambda})\equiv
            A_0(1,\boldsymbol{\lambda}) + U_1(\boldsymbol{\lambda}).

        Args:
            v (array_like): The nondimensional end separation.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`parallel_calculation`.

        Returns:
            numpy.ndarray: The relative nondimensional Helmholtz free energy.

        """
        v = self.np_array(v)
        beta_A = np.zeros(v.shape)
        for i, v_i in enumerate(v):
            self.beta_E = lambda lambda_: self.beta_U_1(lambda_) + \
                self.beta_A_0_abs_isometric(1, lambda_)

            def serial_fun(init_config, **kwargs):
                return self.beta_A_isometric_monte_carlo_serial(
                        v_i, init_config, **kwargs
                    )

            beta_A[i] = self.parallel_calculation(
                serial_fun,
                self.minimize_beta_U(v_i)[2][-self.M:, 0],
                **kwargs
            )
        return beta_A

    def beta_G_isotensional_monte_carlo_serial(self, p, init_config,
                                               num_samples=1000000, **kwargs):
        """Serial calculation for :meth:`beta_G_isotensional_monte_carlo`.

        Args:
            p (array_like): The nondimensional end force.
            init_config (np.ndarray):
                The initial configuration, typically already burned-in.
            num_samples (int, optional, default=1000000):
                The number of samples to use.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`mh_next_config`.

        Returns:
            numpy.ndarray: The relative nondimensional Gibbs free energy.

        """
        exp_neg_beta_Delta_G = 0
        config = init_config
        for counter in range(1, 1 + num_samples):
            exp_neg_beta_Delta_G_next = np.exp(
                -self.beta_G_b_isotensional(p, config)
            )
            exp_neg_beta_Delta_G += (
                exp_neg_beta_Delta_G_next - exp_neg_beta_Delta_G
            )/counter
            config = self.mh_next_config(config, **kwargs)
        return np.log(1/exp_neg_beta_Delta_G)

    def beta_G_isotensional_monte_carlo(self, p, **kwargs):
        r"""The relative nondimensional Gibbs free energy
        as a function of the nondimensional end force
        in the isometric ensemble, using a
        Metropolis-Hastings Markov chain Monte Carlo calculation
        of the ensemble average

        .. math::
            \beta\Delta G(v) =
            -\ln\left\langle e^{-\beta\Delta G_0} \right\rangle_\star,

        where the relative free energy of the reference system is

        .. math::
            \Delta G_0\equiv
            G_0(p,\boldsymbol{\lambda})-G_0(0,\boldsymbol{\lambda}),

        and where the isotensional ensemble average here is

        .. math::
            \left\langle \phi \right\rangle_\star \equiv
            \frac{
                \int d\lambda \ e^{-\beta G_\star(\boldsymbol{\lambda})}
                    \, \phi(\boldsymbol{\lambda})
            }{
                \int d\lambda \ e^{-\beta G_\star(\boldsymbol{\lambda})}
            }
            ,

        where the free energy measure here is

        .. math::
            G_\star(\boldsymbol{\lambda})\equiv
            G_0(0,\boldsymbol{\lambda}) + U_1(\boldsymbol{\lambda}).

        Args:
            p (array_like): The nondimensional end force.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`parallel_calculation`.

        Returns:
            numpy.ndarray: The relative nondimensional Gibbs free energy.

        """
        p = self.np_array(p)
        beta_G = np.zeros(p.shape)
        for i, p_i in enumerate(p):
            self.beta_E = lambda lambda_: self.beta_U_1(lambda_) + \
                self.beta_G_0_abs_isotensional(0, lambda_)

            def serial_fun(init_config, **kwargs):
                return self.beta_G_isotensional_monte_carlo_serial(
                        p_i, init_config, **kwargs
                    )

            beta_G[i] = self.parallel_calculation(
                serial_fun,
                self.minimize_beta_Pi(p_i)[2][-self.M:, 0],
                **kwargs
            )
        return beta_G

    def k_isometric_monte_carlo_serial(self, v, init_config,
                                       num_samples=1000000, **kwargs):
        """Serial calculation for :meth:`k_isometric_monte_carlo`.

        Args:
            v (array_like): The nondimensional end separation.
            init_config (np.ndarray):
                The initial configuration, typically already burned-in.
            num_samples (int, optional, default=1000000):
                The number of samples to use.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`mh_next_config`.

        Returns:
            numpy.ndarray: The nondimensional forward reaction rate.

        """
        exp_neg_beta_Delta_A_0 = 0
        config = init_config
        for counter in range(1, 1 + num_samples):
            if len(init_config) == self.M:
                lambda_ = config
            else:
                lambda_ = np.concatenate(([self.lambda_TS], config))
            exp_neg_beta_Delta_A_0_next = np.exp(
                -self.beta_A_b_isometric(v, lambda_)
            )
            exp_neg_beta_Delta_A_0 += (
                exp_neg_beta_Delta_A_0_next - exp_neg_beta_Delta_A_0
            )/counter
            config = self.mh_next_config(config, **kwargs)
        return exp_neg_beta_Delta_A_0

    def k_isometric_monte_carlo(self, v, **kwargs):
        r"""The nondimensional forward reaction rate coefficient
        as a function of the nondimensional end force
        in the isometric ensemble, using a
        Metropolis-Hastings Markov chain Monte Carlo calculation
        of the ensemble average

        .. math::
            \frac{k(v)}{k(1)} =
                \frac{\left\langle
                    e^{-\beta\Delta A_0^\ddagger}
                \right\rangle_\star^\ddagger}{\left\langle
                    e^{-\beta\Delta A_0}
                \right\rangle_\star}.

        Args:
            v (array_like): The nondimensional end separation.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`parallel_calculation`.

        Returns:
            numpy.ndarray: The nondimensional forward reaction rate.

        """
        v = self.np_array(v)
        ensemble_average_fun = np.zeros(v.shape)
        for i, v_i in enumerate(v):
            self.beta_E = lambda lambda_: self.beta_U_1(lambda_) + \
                self.beta_A_0_abs_isometric(1, lambda_)

            def serial_fun(init_config, **kwargs):
                return self.k_isometric_monte_carlo_serial(
                        v_i, init_config, **kwargs
                    )

            ensemble_average_fun[i] = self.parallel_calculation(
                serial_fun,
                self.minimize_beta_U(v_i)[2][-self.M:, 0],
                **kwargs
            )
        ensemble_average_fun_TS = np.zeros(v.shape)
        for i, v_i in enumerate(v):
            self.beta_E = lambda lambda_: \
                self.beta_U_1(
                    np.concatenate(([self.lambda_TS], lambda_))
                ) + self.beta_A_0_abs_isometric(
                    1, np.concatenate(([self.lambda_TS], lambda_))
                )

            def serial_fun(init_config, **kwargs):
                return self.k_isometric_monte_carlo_serial(
                        v_i, init_config, **kwargs
                    )

            ensemble_average_fun_TS[i] = self.parallel_calculation(
                serial_fun,
                self.minimize_beta_U(
                    v_i, transition_state=True
                )[2][-(self.M - 1):, 0],
                **kwargs
            )
        return ensemble_average_fun_TS/ensemble_average_fun

    def k_isotensional_monte_carlo_serial(self, p, init_config,
                                          num_samples=1000000, **kwargs):
        """Serial calculation for :meth:`k_isotensional_monte_carlo`.

        Args:
            init_config (np.ndarray):
                The initial configuration, typically already burned-in.
            num_samples (int, optional, default=1000000):
                The number of samples to use.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`mh_next_config`.

        Returns:
            numpy.ndarray: The nondimensional forward reaction rate.

        """
        exp_neg_beta_Delta_G_0 = 0
        config = init_config
        for counter in range(1, 1 + num_samples):
            if len(init_config) == self.M:
                lambda_ = config
            else:
                lambda_ = np.concatenate(([self.lambda_TS], config))
            exp_neg_beta_Delta_G_0_next = np.exp(
                -self.beta_G_b_isotensional(p, lambda_)
            )
            exp_neg_beta_Delta_G_0 += (
                exp_neg_beta_Delta_G_0_next - exp_neg_beta_Delta_G_0
            )/counter
            config = self.mh_next_config(config, **kwargs)
        return exp_neg_beta_Delta_G_0

    def k_isotensional_monte_carlo(self, p, **kwargs):
        r"""The nondimensional forward reaction rate coefficient
        as a function of the nondimensional end force
        in the isotensional ensemble, using a
        Metropolis-Hastings Markov chain Monte Carlo calculation
        of the ensemble average

        .. math::
            \frac{k(p)}{k(0)} =
                \frac{\left\langle
                    e^{-\beta\Delta G_0^\ddagger}
                \right\rangle_\star^\ddagger}{\left\langle
                    e^{-\beta\Delta G_0}
                \right\rangle_\star}.

        Args:
            p (array_like): The nondimensional end force.
            **kwargs: Arbitrary keyword arguments.
                Passed to :meth:`parallel_calculation`.

        Returns:
            numpy.ndarray: The nondimensional forward reaction rate.

        """
        p = self.np_array(p)
        ensemble_average_fun = np.zeros(p.shape)
        for i, p_i in enumerate(p):
            self.beta_E = lambda lambda_: self.beta_U_1(lambda_) + \
                self.beta_G_0_abs_isotensional(0, lambda_)

            def serial_fun(init_config, **kwargs):
                return self.k_isotensional_monte_carlo_serial(
                        p_i, init_config, **kwargs
                    )

            ensemble_average_fun[i] = self.parallel_calculation(
                serial_fun,
                self.minimize_beta_Pi(p_i)[2][-self.M:, 0],
                **kwargs
            )
        ensemble_average_fun_TS = np.zeros(p.shape)
        for i, p_i in enumerate(p):
            self.beta_E = lambda lambda_: \
                self.beta_U_1(
                    np.concatenate(([self.lambda_TS], lambda_))
                ) + self.beta_G_0_abs_isotensional(
                    0, np.concatenate(([self.lambda_TS], lambda_))
                )

            def serial_fun(init_config, **kwargs):
                return self.k_isotensional_monte_carlo_serial(
                        p_i, init_config, **kwargs
                    )

            ensemble_average_fun_TS[i] = self.parallel_calculation(
                serial_fun,
                self.minimize_beta_Pi(
                    p_i, transition_state=True
                )[2][-(self.M - 1):, 0],
                **kwargs
            )
        return ensemble_average_fun_TS/ensemble_average_fun



"""A module for basic crack model utilities.

This module consist of the class :class:`BasicUtility` which contains
attributes and methods that are meant to be inherited and utilized as
basic utilities by an arbitrary crack model class.

"""

import sys
import numpy as np
from scipy.optimize import root_scalar


class BasicUtility:
    """The crack model basic utilities class.

    This class contains
    attributes and methods that are meant to be inherited and utilized as
    basic utilities by an arbitrary crack model class.

    Attributes:
        minimum float:
            Helps avoid overflow when dividing by small numbers.
        maximum_float:
            Helps avoid overflow in function evaluations.
        maximum_exponent:
            Helps avoid overflow in exponential functions.

    """
    def __init__(self):
        """Initializes the :class:`BasicUtility` class.

        Inherits overflow-related parameters from ``sys.float_info``.

        """
        self.h = 1e-6
        self.minimum_float = sys.float_info.min
        self.maximum_float = sys.float_info.max
        self.maximum_exponent = np.log(self.maximum_float)

    def np_array(self, input_):
        """Function to return input as a numpy array.

        This function essentially serves as a wrapper for ``numpy.array``
        that returns non-array type inputs as shape (1,) numpy arrays
        rather than shape () numpy arrays, which is useful for indexing.

        Args:
            input_ (array_like): Anything passable to ``numpy.array``.

        Returns:
            numpy.ndarray: The input as an index-able numpy array.

        Example:
            Compare attempts to index the numpy array of an integer:

                >>> import numpy as np
                >>> from statmechcrack.utility import BasicUtility
                >>> try:
                ...     np.array(8)[0]
                ... except IndexError:
                ...     print('Error indexing 0-dimensional array.')
                ... finally:
                ...     BasicUtility().np_array(8)[0]
                Error indexing 0-dimensional array.
                8

        """
        if not np.array(input_).shape:
            return np.array([input_])
        return np.array(input_)

    def inv_fun(self, fun, val, guess=None, **kwargs):
        """Function to invert a mathematical function.

        This function returns the argument x given a function f(x)
        and y, the query value of the function, i.e. y = f(x).

        Note:
            This function is only meant for bijective f(x).

        Args:
            fun (function): The function f(x).
            val (array_like): The value(s) of the function y.
            guess (list, optional): A list of two initial guesses.
            **kwargs: Arbitrary keyword arguments.
                Passed to ``root_scalar``.

        Returns:
            numpy.ndarray: The corresponding argument(s) x.

        """
        vals = self.np_array(val)
        arg = np.zeros(vals.shape)
        for i, val_i in enumerate(vals):
            def res(arg):
                return fun(arg) - val_i
            if guess is not None:
                guess_0 = guess*0.95
                guess_1 = guess
            elif val_i == 0:
                guess_0 = 0
                guess_1 = 1e-2
            else:
                guess_0 = val_i*0.95
                guess_1 = val_i
            arg[i] = root_scalar(res, x0=guess_0, x1=guess_1, **kwargs).root
        return arg



"""A module for testing examples within docstrings.

This module checks all other modules in the package
and ensures that the examples within each docstring
are executed properly and obtain the expected output.

"""

import doctest
import unittest
from os import path
from glob import glob


class Examples(unittest.TestCase):
    """Class to test examples within docstrings.

    """
    def main(self):
        """Main function for module-level testing functionality.

        """
        self.test_docstring_examples()

    def test_docstring_examples(self):
        """Function to test examples within docstrings.

        """
        tests_dir = path.dirname(__file__)
        files = glob(path.join(tests_dir, './*.py'))
        files += glob(path.join(tests_dir, '../*.py'))
        for file in files:
            self.assertFalse(doctest.testfile(file, module_relative=False)[0])


if __name__ == '__main__':
    unittest.main()



"""Module-level functionality for testing installation.

Example:
    Test that the package was installed properly:

    ::

        python -m statmechcrack.tests

    The example and style testing can also be included:

    ::

        python -m statmechcrack.tests -x

"""

import argparse

from .test_zero import Zero
from .test_one import One
from .test_many import Many
from .test_boundary import Boundary
from .test_interface import Interface

Zero().main()
One().main()
Many().main()
Boundary().main()
Interface().main()

parser = argparse.ArgumentParser()
parser.add_argument('-x', action='store_true')
args = parser.parse_args()
if parser.parse_args().x is True:
    from .test_examples import Examples
    from .test_style import Style
    Examples().main()
    Style().main()



"""A module for testing when something is one.

"""

import unittest
import numpy as np

from ..utility import BasicUtility
from .test_zero import random_crack_model


class One(unittest.TestCase):
    """Class to test for when something is one.

    """
    def main(self):
        """Main function for module-level testing functionality.

        """
        self.test_one_inverse()
        self.test_one_nondimensional_end_separation()
        self.test_one_relative_nondimensional_reaction_rate_coefficient()

    def test_one_inverse(self):
        """Function to test inverse calculation at one.

        """
        self.assertEqual(BasicUtility().inv_fun(lambda x: x, 1), 1)

    def test_one_nondimensional_end_separation(self):
        """Function to test for a nondimensional end separation of one.

        """
        rgn = np.random.rand()
        model = random_crack_model()
        self.assertAlmostEqual(
            model.v_b_mechanical(
                0, [rgn, rgn]
            )[0], rgn
        )
        self.assertAlmostEqual(
            model.v_0_mechanical(
                0, [rgn, rgn]
            )[0], rgn
        )
        self.assertAlmostEqual(
            model.v_mechanical(
                0
            )[0], 1
        )
        self.assertAlmostEqual(
            model.v_b_isotensional(
                0, [rgn, rgn]
            )[0], rgn
        )
        self.assertAlmostEqual(
            model.v_0_isotensional(
                0, [rgn, rgn]
            )[0], rgn
        )
        self.assertAlmostEqual(
            model.v_isotensional(
                0, approach='asymptotic'
            )[0], 1
        )
        self.assertAlmostEqual(
            model.v_b(
                0, [rgn, rgn], ensemble='isometric'
            )[0], rgn
        )
        self.assertAlmostEqual(
            model.v_b(
                0, [rgn, rgn], ensemble='isotensional'
            )[0], rgn
        )
        self.assertAlmostEqual(
            model.v_0(
                0, [rgn, rgn], ensemble='isometric'
            )[0], rgn
        )
        self.assertAlmostEqual(
            model.v_0(
                0, [rgn, rgn], ensemble='isotensional'
            )[0], rgn
        )
        self.assertAlmostEqual(
            model.v(
                0, ensemble='isometric', approach='asymptotic'
            )[0], 1
        )
        self.assertAlmostEqual(
            model.v(
                0, ensemble='isotensional', approach='asymptotic'
            )[0], 1
        )

    def test_one_relative_nondimensional_reaction_rate_coefficient(self):
        """Function to test for a relative nondimensional
        reaction rate coefficient of one.

        """
        rgn0, rgn1 = np.random.rand(2)
        model = random_crack_model()
        self.assertEqual(
            model.k_b_isometric(
                1, [rgn0, rgn1]
            )[0], 1
        )
        self.assertEqual(
            model.k_0_isometric(
                1, [rgn0, rgn1]
            )[0], 1
        )
        self.assertEqual(
            model.k_isometric(
                1, approach='asymptotic'
            )[0], 1
        )
        self.assertEqual(
            model.k_b_isotensional(
                0, [rgn0, rgn1]
            )[0], 1
        )
        self.assertEqual(
            model.k_0_isotensional(
                0, [rgn0, rgn1]
            )[0], 1
        )
        self.assertEqual(
            model.k_isotensional(
                0, approach='asymptotic'
            )[0], 1
        )
        self.assertEqual(
            model.k_b(
                1, [rgn0, rgn1], ensemble='isometric'
            )[0], 1
        )
        self.assertEqual(
            model.k_0(
                1, [rgn0, rgn1], ensemble='isometric'
            )[0], 1
        )
        self.assertEqual(
            model.k(
                1, ensemble='isometric', approach='asymptotic'
            )[0], 1
        )
        self.assertEqual(
            model.k_b(
                0, [rgn0, rgn1], ensemble='isotensional'
            )[0], 1
        )
        self.assertEqual(
            model.k_0(
                0, [rgn0, rgn1], ensemble='isotensional'
            )[0], 1
        )
        self.assertEqual(
            model.k(
                0, ensemble='isotensional', approach='asymptotic'
            )[0], 1
        )


if __name__ == '__main__':
    unittest.main()



"""A module for testing the Python style.

This module checks all other modules in the package (including itself)
for adherence to the PEP 8 style standard for Python code.

"""

import unittest
from os import path
from glob import glob
from pycodestyle import StyleGuide


class Style(unittest.TestCase):
    """Class to test conformance of code to PEP style standards.

    """
    def main(self):
        """Main function for module-level testing functionality.

        """
        self.test_pep8_conformance()

    def test_pep8_conformance(self):
        """Function to test conformance of code to PEP style standards.

        """
        tests_dir = path.dirname(__file__)
        files = glob(path.join(tests_dir, './*.py'))
        files += glob(path.join(tests_dir, '../*.py'))
        files += glob(path.join(tests_dir, '../../*.py'))
        files += glob(path.join(tests_dir, '../../docs/*.py'))
        style = StyleGuide(quiet=False)
        for file in files:
            self.assertFalse(style.input_file(file))


if __name__ == '__main__':
    unittest.main()
